#
# LMirror is Copyright (C) 2010 Robert Collins <robertc@robertcollins.net>
# 
# LMirror is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
# 
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE.  See the GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <http://www.gnu.org/licenses/>.
# 
# In the LMirror source tree the file COPYING.txt contains the GNU General Public
# License version 3.
# 

"""Tests for the commands module."""

import os.path
import sys

from testresources import TestResource

from l_mirror import commands
from l_mirror.tests import ResourcedTestCase
from l_mirror.tests.matchers import MatchesException
from l_mirror.tests.stubpackage import (
    StubPackageResource,
    )
from l_mirror.ui import cli, model


class TemporaryCommand(object):
    """A temporary command."""


class TemporaryCommandResource(TestResource):

    def __init__(self, cmd_name):
        TestResource.__init__(self)
        self.resources.append(('pkg',
            StubPackageResource('commands',
            [('%s.py' % cmd_name,
             """from l_mirror.commands import Command
class %s(Command):
    def run(self):
        pass
""" % cmd_name)], init=False)))
        self.cmd_name = cmd_name

    def make(self, dependency_resources):
        pkg = dependency_resources['pkg']
        result = TemporaryCommand()
        result.path = os.path.join(pkg.base, 'commands')
        commands.__path__.append(result.path)
        return result

    def clean(self, resource):
        commands.__path__.remove(resource.path)
        name = 'l_mirror.commands.%s' % self.cmd_name
        if name in sys.modules:
            del sys.modules[name]


class TestFindCommand(ResourcedTestCase):

    resources = [('cmd', TemporaryCommandResource('foo'))]

    def test_looksupcommand(self):
        cmd = commands._find_command('foo')
        self.assertIsInstance(cmd(None), commands.Command)

    def test_missing_command(self):
        self.assertRaises(KeyError, commands._find_command, 'bar')

    def test_sets_name(self):
        cmd = commands._find_command('foo')
        self.assertEqual('foo', cmd.name)


class TestFindHyphenCommand(ResourcedTestCase):

    resources = [('cmd', TemporaryCommandResource('foo_bar'))]

    def test_looksupcommand_hypen(self):
        cmd = commands._find_command('foo-bar')
        self.assertIsInstance(cmd(None), commands.Command)


class TestIterCommands(ResourcedTestCase):

    resources = [
        ('cmd1', TemporaryCommandResource('one')),
        ('cmd2', TemporaryCommandResource('two')),
        ]

    def test_iter_commands(self):
        cmds = list(commands.iter_commands())
        cmds = [cmd(None).name for cmd in cmds]
        # We don't care about all the built in commands
        cmds = [cmd for cmd in cmds if cmd in ('one', 'two')]
        self.assertEqual(['one', 'two'], cmds)


class InstrumentedCommand(commands.Command):
    """A command which records methods called on it.
    
    The first line is the summary.
    """

    def _init(self):
        self.calls = []

    def execute(self):
        self.calls.append('execute')
        return commands.Command.execute(self)

    def run(self):
        self.calls.append('run')


class TestAbstractCommand(ResourcedTestCase):

    def test_execute_calls_run(self):
        cmd = InstrumentedCommand(model.UI())
        self.assertEqual(0, cmd.execute())
        self.assertEqual(['execute', 'run'], cmd.calls)

    def test_execute_calls_set_command(self):
        ui = model.UI()
        cmd = InstrumentedCommand(ui)
        cmd.execute()
        self.assertEqual(cmd, ui.cmd)

    def test_execute_does_not_run_if_set_command_errors(self):
        class FailUI(object):
            def set_command(self, ui):
                return False
        cmd = InstrumentedCommand(FailUI())
        self.assertEqual(1, cmd.execute())

    def test_shows_errors_from_execute_returns_3(self):
        class FailCommand(commands.Command):
            def run(self):
                raise Exception("foo")
        ui = model.UI()
        cmd = FailCommand(ui)
        self.assertEqual(3, cmd.execute())
        self.assertEqual(1, len(ui.outputs))
        self.assertEqual('error', ui.outputs[0][0])
        self.assertThat(ui.outputs[0][1], MatchesException(Exception('foo')))

    def test_get_summary(self):
        cmd = InstrumentedCommand
        self.assertEqual('A command which records methods called on it.',
            cmd.get_summary())



#
# LMirror is Copyright (C) 2010 Robert Collins <robertc@robertcollins.net>
# 
# LMirror is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
# 
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE.  See the GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <http://www.gnu.org/licenses/>.
# 
# In the LMirror source tree the file COPYING.txt contains the GNU General Public
# License version 3.
# 

"""Tests for matchers used by or for testing l_mirror."""

import sys

from l_mirror.tests import ResourcedTestCase
from l_mirror.tests.matchers import MatchesException


class TestMatchesException(ResourcedTestCase):

    def test_does_not_match_different_exception_class(self):
        matcher = MatchesException(ValueError("foo"))
        try:
            raise Exception("foo")
        except Exception:
            error = sys.exc_info()
        mismatch = matcher.match(error)
        self.assertNotEqual(None, mismatch)
        self.assertEqual(
            "<type 'exceptions.Exception'> is not a "
            "<type 'exceptions.ValueError'>",
            mismatch.describe())

    def test_does_not_match_different_args(self):
        matcher = MatchesException(Exception("foo"))
        try:
            raise Exception("bar")
        except Exception:
            error = sys.exc_info()
        mismatch = matcher.match(error)
        self.assertNotEqual(None, mismatch)
        self.assertEqual(
            "Exception('bar',) has different arguments to Exception('foo',).",
            mismatch.describe())

    def test_matches_same_args(self):
        matcher = MatchesException(Exception("foo"))
        try:
            raise Exception("foo")
        except Exception:
            error = sys.exc_info()
        mismatch = matcher.match(error)
        self.assertEqual(None, mismatch)



#
# LMirror is Copyright (C) 2010 Robert Collins <robertc@robertcollins.net>
# 
# LMirror is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
# 
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE.  See the GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <http://www.gnu.org/licenses/>.
# 
# In the LMirror source tree the file COPYING.txt contains the GNU General Public
# License version 3.
# 

"""Initialise a mirror set."""

from optparse import Option

from bzrlib import urlutils

from l_mirror.arguments import path
from l_mirror.commands import Command
from l_mirror import mirrorset

class init(Command):
    """Initialise a mirror set.
    
    Take the place and name to create the mirror set, and optionally a 
    different directory to replicate. Note that if the directory to
    replicate is the same as the location of the mirror set configuration, then
    config changes will automatically propogate - this is the default if you
    just run (for example)::
    
      $ lmirror init /srv/ubuntu/ubuntu
    
    which will create a mirror set called ubuntu in /srv/ubuntu.

    Creating a mirror set does a scan-and-hash of the entire contents of that
    directory, so it will take time run about the same as tarring up the entire
    directory.
    """

    args = [path.PathArgument('mirror_set', min=1, max=1),
        path.PathArgument('content_root', min=0)]
    options = [Option("--empty", dest="empty", help="Create the new mirror set"
        " as empty. This is useful when you intend to customise and tuen the"
        " filters for it and don't want to wait for an initial scan to take"
        " place.", action="store_true", default=False)
        ]

    def run(self):
        transport = self.ui.arguments['mirror_set'][0]
        base = transport.clone('..')
        name = base.relpath(transport.base)
        if not self.ui.arguments['content_root']:
            content_root = base
        else:
            content_root = self.ui.arguments['content_root'][0]
        mirror = mirrorset.initialise(base, name, content_root, self.ui)
        if not self.ui.options.empty:
            mirror.finish_change()
        return 0



#
# LMirror is Copyright (C) 2010 Robert Collins <robertc@robertcollins.net>
# 
# LMirror is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
# 
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE.  See the GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <http://www.gnu.org/licenses/>.
# 
# In the LMirror source tree the file COPYING.txt contains the GNU General Public
# License version 3.
# 

"""Tests for the finish-change command."""

from doctest import ELLIPSIS

from bzrlib.transport import get_transport
from bzrlib.transport.memory import MemoryServer

from testtools.matchers import DocTestMatches

from l_mirror.commands import finish_change
from l_mirror import mirrorset
from l_mirror.ui.model import UI
from l_mirror.tests import ResourcedTestCase
from l_mirror.tests.matchers import MatchesException


class TestCommandFinishChange(ResourcedTestCase):

    def get_test_ui_and_cmd(self, args):
        ui = UI(args=args)
        cmd = finish_change.finish_change(ui)
        ui.set_command(cmd)
        return ui, cmd

    def setup_memory(self):
        """Create a memory url server and return its url."""
        # XXX: integrate with ui.here better.
        self.transport_factory = MemoryServer()
        self.transport_factory.start_server()
        self.addCleanup(self.transport_factory.stop_server)
        return self.transport_factory.get_url()

    def test_not_updating_errors(self):
        base = self.setup_memory()
        t = get_transport(base)
        t = t.clone('path')
        t.create_prefix()
        root = t.base + 'myname'
        ui, cmd = self.get_test_ui_and_cmd((root,))
        mirror = mirrorset.initialise(t, 'myname', t, UI())
        mirror.finish_change()
        self.assertEqual(3, cmd.execute())
        self.assertEqual(1, len(ui.outputs))
        self.assertEqual('error', ui.outputs[0][0])
        self.assertThat(ui.outputs[0][1],
            MatchesException(ValueError('No changeset open')))

    def test_updating_finishes(self):
        base = self.setup_memory()
        root = base + 'path/myname'
        t = get_transport(base)
        t = t.clone('path')
        contentdir = t
        contentdir.create_prefix()
        contentdir.mkdir('dir1')
        contentdir.mkdir('dir2')
        contentdir.put_bytes('abc', '1234567890\n')
        contentdir.put_bytes('dir1/def', 'abcdef')
        ui, cmd = self.get_test_ui_and_cmd((root,))
        mirror = mirrorset.initialise(t, 'myname', t, ui)
        self.assertEqual(0, cmd.execute())
        t = t.clone('.lmirror/metadata/myname')
        self.assertThat(t.get_bytes('metadata.conf'), DocTestMatches("""[metadata]
basis = 0
latest = 1
timestamp = ...
updating = False

""", ELLIPSIS))
        self.assertThat(t.get_bytes('journals/0'), DocTestMatches("""l-mirror-journal-1
"""))
        self.assertThat(t.get_bytes('journals/1'), DocTestMatches("""l-mirror-journal-1
.lmirror\x00new\x00dir\x00.lmirror/sets\x00new\x00dir\x00.lmirror/sets/myname\x00new\x00dir\x00.lmirror/sets/myname/format\x00new\x00file\x00e5fa44f2b31c1fb553b6021e7360d07d5d91ff5e\x002\x00.lmirror/sets/myname/set.conf\x00new\x00file\x00061df21cf828bb333660621c3743cfc3a3b2bd23\x0023\x00abc\x00new\x00file\x0012039d6dd9a7e27622301e935b6eefc78846802e\x0011\x00dir1\x00new\x00dir\x00dir1/def\x00new\x00file\x001f8ac10f23c5b5bc1167bda84b833e5c057a77d2\x006\x00dir2\x00new\x00dir"""))



#
# LMirror is Copyright (C) 2010 Robert Collins <robertc@robertcollins.net>
# 
# LMirror is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
# 
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE.  See the GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <http://www.gnu.org/licenses/>.
# 
# In the LMirror source tree the file COPYING.txt contains the GNU General Public
# License version 3.
# 

"""Mirror an existing mirror set."""

from bzrlib import urlutils

from l_mirror.arguments import path, url
from l_mirror.commands import Command
from l_mirror import mirrorset

class mirror(Command):
    """Mirror a mirror set.

    Takes the mirror to mirror from, and the mirror to mirror too. If the
    target does not exist it will be populated from the source mirror.
    
    For example:
    
      $ lmirror mirror http://archive.ubuntu.com/ubuntu/ubuntu /srv/ubuntu

    will initialise a mirror set called ubuntu in /srv/ubuntu, with 
    http://archive.ubuntu.com/ubuntu/ as a known sender.
    
    Mirroring will not generally copy content already present, but the first
    mirror run can still take considerable time. Subsequent runs should be
    extremely fast.
    """

    args = [url.URLArgument('source_mirror', min=1, max=1),
        path.PathArgument('target_mirror', min=1, max=1)]

    def run(self):
        source_transport = self.ui.arguments['source_mirror'][0]
        source_base = source_transport.clone('..')
        name = source_base.relpath(source_transport.base)
        target_base = self.ui.arguments['target_mirror'][0]
        source = mirrorset.MirrorSet(source_base, name, self.ui)
        try:
            target = mirrorset.MirrorSet(target_base, name, self.ui)
        except ValueError:
            # No mirror there yet
            target = mirrorset.initialise(target_base, name,
                target_base.clone(source.content_root_path()), self.ui)
            target.cancel_change()
        target.receive(source)
        return 0



#
# LMirror is Copyright (C) 2010 Robert Collins <robertc@robertcollins.net>
# 
# LMirror is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
# 
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE.  See the GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <http://www.gnu.org/licenses/>.
# 
# In the LMirror source tree the file COPYING.txt contains the GNU General Public
# License version 3.
# 

"""Useful generic logging config stuff, mainly used by l_mirror.ui.cl."""

__all__ = ['configure_logging']

import logging
import os.path
import time

from bzrlib import transport

def configure_logging(console_stream, f_path="~/.cache/lmirror/log"):
    """Configure a formatter and create a console and file handlers.

    The formatter is setup for ISO8601 UTC logging.

    :param console_stream: A stream to log to for console events.
    :param f_path: Optional path to configure for the file logging (used in
        testing). Use None to disable file logging.
    :return (console_log_handler, file_log_handler, formatter).
    """
    f_formatter = logging.Formatter("%(asctime)s: %(message)s", "%Y-%m-%d %H:%M:%SZ")
    f_formatter.converter = time.gmtime
    console_handler = logging.StreamHandler(console_stream)
    # console_handler.setFormatter(formatter)
    # Capture root events
    logger = logging.getLogger()
    logger.addHandler(console_handler)
    logger.setLevel(0)
    if f_path is None:
        file_handler = None
    else:
        f_path = os.path.expanduser(f_path)
        if not os.path.exists(f_path):
            t = transport.get_transport(f_path).clone('..')
            t.create_prefix()
        file_handler = logging.FileHandler(f_path)
        file_handler.setFormatter(f_formatter)
        logger.addHandler(file_handler)
    return console_handler, file_handler, f_formatter

    return None, None, None



#
# LMirror is Copyright (C) 2010 Robert Collins <robertc@robertcollins.net>
# 
# LMirror is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
# 
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE.  See the GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <http://www.gnu.org/licenses/>.
# 
# In the LMirror source tree the file COPYING.txt contains the GNU General Public
# License version 3.
# 

"""The l_mirror library.

This library is divided into some broad areas.

The arguments package contains command argument parsing facilities.
The commands package contains the main user entry points into the application.
The ui package contains various user interfaces.
The journals module contains logic for working with individual journals.
The mirrorset module contains logic for working with a mirrorset - a thing to
    be mirrored around.
The tests package contains tests and test specific support code.
"""

__all__ = ['__version__']

# same format as sys.version_info: "A tuple containing the five components of
# the version number: major, minor, micro, releaselevel, and serial. All
# values except releaselevel are integers; the release level is 'alpha',
# 'beta', 'candidate', or 'final'. The version_info value corresponding to the
# Python version 2.0 is (2, 0, 0, 'final', 0)."  Additionally we use a
# releaselevel of 'dev' for unreleased under-development code.
#
# If the releaselevel is 'alpha' then the major/minor/micro components are not
# established at this point, and setup.py will use a version of next-$(revno).
# If the releaselevel is 'final', then the tarball will be major.minor.micro.
# Otherwise it is major.minor.micro~$(revno).
__version__ = (0, 0, 1, 'final', 0)



#
# LMirror is Copyright (C) 2010 Robert Collins <robertc@robertcollins.net>
# 
# LMirror is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
# 
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE.  See the GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <http://www.gnu.org/licenses/>.
# 
# In the LMirror source tree the file COPYING.txt contains the GNU General Public
# License version 3.
# 

"""Tests for the mirrorset module."""

from doctest import ELLIPSIS

from bzrlib import gpg as bzrgpg
from bzrlib.transport import get_transport
from bzrlib.transport.memory import MemoryServer

from testtools.matchers import DocTestMatches

from l_mirror import gpg, mirrorset
from l_mirror.ui.model import UI
from l_mirror.tests import ResourcedTestCase


class TestMirrorSet(ResourcedTestCase):

    def get_test_ui(self):
        ui = UI()
        return ui

    def setup_memory(self):
        """Create a memory url server and return its url."""
        # XXX: integrate with ui.here better.
        self.transport_factory = MemoryServer()
        self.transport_factory.start_server()
        self.addCleanup(self.transport_factory.stop_server)
        return self.transport_factory.get_url()

    def test_create_set(self):
        basedir = get_transport(self.setup_memory()).clone('path')
        basedir.create_prefix()
        ui = self.get_test_ui()
        mirror = mirrorset.initialise(basedir, 'myname', basedir, ui)
        t = basedir.clone('.lmirror/sets/myname')
        self.assertEqual('1\n', t.get_bytes('format'))
        self.assertThat(t.get_bytes('set.conf'), DocTestMatches("""[set]
content_root = .
"""))
        self.assertEqual(basedir, mirror.base)
        self.assertEqual('myname', mirror.name)
        self.assertEqual(ui, mirror.ui)

    def test_error_creating_existing_set(self):
        basedir = get_transport(self.setup_memory()).clone('path')
        basedir.create_prefix()
        ui = self.get_test_ui()
        mirror = mirrorset.initialise(basedir, 'myname', basedir, ui)
        self.assertRaises(ValueError,
            mirrorset.initialise, basedir, 'myname', basedir, ui)

    def test_accepts_content_root(self):
        basedir = get_transport(self.setup_memory()).clone('path')
        basedir.create_prefix()
        contentdir = basedir.clone('../content')
        contentdir.create_prefix()
        ui = self.get_test_ui()
        mirror = mirrorset.initialise(basedir, 'myname', contentdir, ui)
        t = basedir.clone('.lmirror/sets/myname')
        self.assertEqual('1\n', t.get_bytes('format'))
        self.assertThat(t.get_bytes('set.conf'), DocTestMatches("""[set]
content_root = ../content
"""))

    def test_second_set_can_be_created(self):
        basedir = get_transport(self.setup_memory()).clone('path')
        basedir.create_prefix()
        ui = self.get_test_ui()
        mirror = mirrorset.initialise(basedir, 'myname', basedir, ui)
        mirror = mirrorset.initialise(basedir, 'myname2', basedir, ui)
        t = basedir.clone('.lmirror/sets/myname2')
        self.assertEqual('1\n', t.get_bytes('format'))
        self.assertThat(t.get_bytes('set.conf'), DocTestMatches("""[set]
content_root = .
"""))

    def test_creates_metadata_dir(self):
        basedir = get_transport(self.setup_memory()).clone('path')
        basedir.create_prefix()
        ui = self.get_test_ui()
        mirror = mirrorset.initialise(basedir, 'myname', basedir, ui)
        t = basedir.clone('.lmirror/metadata/myname')
        self.assertEqual('1\n', t.get_bytes('format'))
        self.assertThat(t.get_bytes('metadata.conf'), DocTestMatches("""[metadata]
basis = 0
latest = 0
timestamp = 0
updating = True
""", ELLIPSIS))
        self.assertThat(t.get_bytes('journals/0'), DocTestMatches("""l-mirror-journal-1
"""))

    def test_open_set(self):
        basedir = get_transport(self.setup_memory()).clone('path')
        basedir.create_prefix()
        ui = self.get_test_ui()
        mirrorset.initialise(basedir, 'myname', basedir, ui)
        t = basedir.clone('.lmirror/sets/myname')
        self.assertEqual('1\n', t.get_bytes('format'))
        self.assertThat(t.get_bytes('set.conf'), DocTestMatches("""[set]
content_root = .
"""))   
        mirror = mirrorset.MirrorSet(basedir, 'myname', ui)

    def test_open_missing(self):
        basedir = get_transport(self.setup_memory()).clone('path')
        basedir.create_prefix()
        ui = self.get_test_ui()
        self.assertRaises(ValueError, mirrorset.MirrorSet, basedir, 'myname', ui)

    def test_open_wrong_format(self):
        basedir = get_transport(self.setup_memory()).clone('path')
        basedir.create_prefix()
        ui = self.get_test_ui()
        mirrorset.initialise(basedir, 'myname', basedir, ui)
        t = basedir.clone('.lmirror/sets/myname')
        self.assertEqual('1\n', t.get_bytes('format'))
        t.put_bytes('format', '2\n')
        self.assertThat(t.get_bytes('set.conf'), DocTestMatches("""[set]
content_root = .
"""))   
        self.assertRaises(ValueError, mirrorset.MirrorSet, basedir, 'myname', ui)

    def test_parse_content_conf(self):
        basedir = get_transport(self.setup_memory()).clone('path')
        basedir.create_prefix()
        ui = self.get_test_ui()
        mirror = mirrorset.initialise(basedir, 'myname', basedir, ui)
        t = basedir.clone('.lmirror/sets/myname')
        t.put_bytes('content.conf', """include a regex
exclude another regex
# a comment
""")
        mirror._parse_content_conf()
        self.assertEqual(['another regex'], mirror.excludes)
        self.assertEqual(['a regex'], mirror.includes)

    def test_start_change_updating_error(self):
        basedir = get_transport(self.setup_memory()).clone('path')
        basedir.create_prefix()
        ui = self.get_test_ui()
        mirror = mirrorset.initialise(basedir, 'myname', basedir, ui)
        self.assertRaises(ValueError, mirror.start_change)

    def test_start_change_not_updating_starts_change(self):
        basedir = get_transport(self.setup_memory()).clone('path')
        basedir.create_prefix()
        ui = self.get_test_ui()
        mirror = mirrorset.initialise(basedir, 'myname', basedir, ui)
        mirror.finish_change()
        mirror.start_change()
        # We know finish_change errors if a change isn't open, so it working is
        # sufficient.
        mirror.finish_change()

    def test_finish_change_not_updating_error(self):
        basedir = get_transport(self.setup_memory()).clone('path')
        basedir.create_prefix()
        ui = self.get_test_ui()
        mirror = mirrorset.initialise(basedir, 'myname', basedir, ui)
        mirror.finish_change()
        self.assertRaises(ValueError, mirror.finish_change)

    def test_finish_change_no_change_no_new_journal(self):
        basedir = get_transport(self.setup_memory()).clone('path')
        basedir.create_prefix()
        ui = self.get_test_ui()
        mirror = mirrorset.initialise(basedir, 'myname', basedir, ui)
        mirror.finish_change()
        mirror.start_change()
        mirror.finish_change()
        self.assertEqual('1', mirror._get_metadata().get('metadata', 'latest'))
        self.assertEqual(('rest', 'No changes found in mirrorset.'),
            ui.outputs[-1])

    def test_finish_change_scans_content(self):
        basedir = get_transport(self.setup_memory()).clone('path')
        basedir.create_prefix()
        ui = self.get_test_ui()
        mirror = mirrorset.initialise(basedir, 'myname', basedir, ui)
        basedir.create_prefix()
        basedir.mkdir('dir1')
        basedir.mkdir('dir2')
        basedir.put_bytes('abc', '1234567890\n')
        basedir.put_bytes('dir1/def', 'abcdef')
        mirror.finish_change()
        t = basedir.clone('.lmirror/metadata/myname')
        self.assertThat(t.get_bytes('metadata.conf'), DocTestMatches("""[metadata]
basis = 0
latest = 1
timestamp = ...
updating = False

""", ELLIPSIS))
        self.assertThat(t.get_bytes('journals/1'), DocTestMatches("""l-mirror-journal-1
.lmirror\x00new\x00dir\x00.lmirror/sets\x00new\x00dir\x00.lmirror/sets/myname\x00new\x00dir\x00.lmirror/sets/myname/format\x00new\x00file\x00e5fa44f2b31c1fb553b6021e7360d07d5d91ff5e\x002\x00.lmirror/sets/myname/set.conf\x00new\x00file\x00061df21cf828bb333660621c3743cfc3a3b2bd23\x0023\x00abc\x00new\x00file\x0012039d6dd9a7e27622301e935b6eefc78846802e\x0011\x00dir1\x00new\x00dir\x00dir1/def\x00new\x00file\x001f8ac10f23c5b5bc1167bda84b833e5c057a77d2\x006\x00dir2\x00new\x00dir"""))
    
    def test_include_excludes_honoured(self):
        basedir = get_transport(self.setup_memory()).clone('path')
        basedir.create_prefix()
        ui = self.get_test_ui()
        mirror = mirrorset.initialise(basedir, 'myname', basedir, ui)
        basedir.create_prefix()
        basedir.mkdir('dir1')
        basedir.mkdir('dir2')
        basedir.put_bytes('abc', '1234567890\n')
        basedir.put_bytes('dir2/included', '1234567890\n')
        basedir.put_bytes('dir1/def', 'abcdef')
        mirror.includes = ['included']
        mirror.excludes = ['dir2/', 'dir1']
        mirror.finish_change()
        t = basedir.clone('.lmirror/metadata/myname')
        self.assertThat(t.get_bytes('metadata.conf'), DocTestMatches("""[metadata]
basis = 0
latest = 1
timestamp = ...
updating = False

""", ELLIPSIS))
        self.assertThat(t.get_bytes('journals/1'), DocTestMatches("""l-mirror-journal-1
.lmirror\x00new\x00dir\x00.lmirror/sets\x00new\x00dir\x00.lmirror/sets/myname\x00new\x00dir\x00.lmirror/sets/myname/format\x00new\x00file\x00e5fa44f2b31c1fb553b6021e7360d07d5d91ff5e\x002\x00.lmirror/sets/myname/set.conf\x00new\x00file\x00061df21cf828bb333660621c3743cfc3a3b2bd23\x0023\x00abc\x00new\x00file\x0012039d6dd9a7e27622301e935b6eefc78846802e\x0011\x00dir2\x00new\x00dir\x00dir2/included\x00new\x00file\x0012039d6dd9a7e27622301e935b6eefc78846802e\x0011"""))
    
    def test_signs_when_there_is_a_keyring(self):
        basedir = get_transport(self.setup_memory()).clone('path')
        basedir.create_prefix()
        ui = self.get_test_ui()
        mirror = mirrorset.initialise(basedir, 'myname', basedir, ui)
        mirror.gpg_strategy = bzrgpg.LoopbackGPGStrategy(None)
        t = basedir.clone('.lmirror/sets/myname')
        t.put_bytes('lmirror.gpg', '')
        mirror.finish_change()
        metadatadir = mirror._metadata()
        self.assertEqual(
            "-----BEGIN PSEUDO-SIGNED CONTENT-----\n" +
            metadatadir.get_bytes('journals/1') +
            "-----END PSEUDO-SIGNED CONTENT-----\n",
            metadatadir.get_bytes('journals/1.sig'))

    def test_receive_replays_and_updates_metadata(self):
        basedir = get_transport(self.setup_memory()).clone('path')
        basedir.create_prefix()
        ui = self.get_test_ui()
        # two journals exist from this simple operation - 0 and 1, but 
        # need 3, as new clones start with 0
        mirror = mirrorset.initialise(basedir, 'myname', basedir, ui)
        basedir.create_prefix()
        basedir.mkdir('dir1')
        basedir.mkdir('dir2')
        basedir.put_bytes('abc', '1234567890\n')
        basedir.put_bytes('dir1/def', 'abcdef')
        mirror.finish_change()
        mirror.start_change()
        basedir.put_bytes('dada', '123456789a\n')
        mirror.finish_change()
        clonedir = basedir.clone('../clone')
        clonedir.create_prefix()
        clone = mirrorset.initialise(clonedir, 'myname', clonedir, ui)
        clone.cancel_change()
        clone.receive(mirror)
        mirrormeta = mirror._get_metadata()
        metadata = clone._get_metadata()
        self.assertEqual('2', metadata.get('metadata', 'latest'))
        self.assertEqual(
            mirrormeta.get('metadata', 'timestamp'),
            metadata.get('metadata', 'timestamp'))
        # check we got a file from each journal
        self.assertEqual('123456789a\n', clonedir.get_bytes('dada'))
        self.assertEqual('1234567890\n', clonedir.get_bytes('abc'))
        # And the journals should be identical.
        mirrorjournal = mirror._journaldir()
        clonejournal = clone._journaldir()
        self.assertEqual(mirrorjournal.get_bytes('1'), clonejournal.get_bytes('1'))
        self.assertEqual(mirrorjournal.get_bytes('2'), clonejournal.get_bytes('2'))

    def test_checks_when_there_is_a_keyring(self):
        basedir = get_transport(self.setup_memory()).clone('path')
        basedir.create_prefix()
        ui = self.get_test_ui()
        mirror = mirrorset.initialise(basedir, 'myname', basedir, ui)
        mirror.gpg_strategy = bzrgpg.LoopbackGPGStrategy(None)
        t = basedir.clone('.lmirror/sets/myname')
        t.put_bytes('lmirror.gpg', '')
        mirror.finish_change()
        metadatadir = mirror._metadata()
        self.assertEqual(
            "-----BEGIN PSEUDO-SIGNED CONTENT-----\n" +
            metadatadir.get_bytes('journals/1') +
            "-----END PSEUDO-SIGNED CONTENT-----\n",
            metadatadir.get_bytes('journals/1.sig'))
        clonedir = basedir.clone('../clone')
        clonedir.create_prefix()
        clone = mirrorset.initialise(clonedir, 'myname', clonedir, ui)
        clone.cancel_change()
        clone.gpgv_strategy = gpg.TestGPGVStrategy([metadatadir.get_bytes('journals/1')])
        clone.receive(mirror)
        metadata = clone._get_metadata()
        self.assertEqual('1', metadata.get('metadata', 'latest'))
        self.assertEqual(
            "-----BEGIN PSEUDO-SIGNED CONTENT-----\n" +
            metadatadir.get_bytes('journals/1') +
            "-----END PSEUDO-SIGNED CONTENT-----\n",
            clone._metadata().get_bytes('journals/1.sig'))



#
# LMirror is Copyright (C) 2010 Robert Collins <robertc@robertcollins.net>
# 
# LMirror is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
# 
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE.  See the GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <http://www.gnu.org/licenses/>.
# 
# In the LMirror source tree the file COPYING.txt contains the GNU General Public
# License version 3.
# 

"""The l_mirror tests and test only code.

Tests are structured to mirror the main code structure. That is, tests for
a module l_mirror.foo.bar are in a test module l_mirrr.tests.foo.test_bar.
"""

import unittest

import testresources
from testscenarios import generate_scenarios
from testtools import TestCase

class ResourcedTestCase(TestCase, testresources.ResourcedTestCase):
    """Make all l_mirror tests have resource support."""

    def setUp(self):
        TestCase.setUp(self)
        testresources.ResourcedTestCase.setUpResources(self)
        self.addCleanup(testresources.ResourcedTestCase.tearDownResources,
            self)


def test_suite():
    packages = [
        'arguments',
        'commands',
        'ui',
        ]
    names = [
        'arguments',
        'commands',
        'journals',
        'logging_resource',
        'logging_support',
        'matchers',
        'mirrorset',
        'monkeypatch',
        'ui',
        'setup',
        'stubpackage',
        'stubhome',
        ]
    module_names = ['l_mirror.tests.test_' + name for name in names]
    loader = unittest.TestLoader()
    suite = loader.loadTestsFromNames(module_names)
    result = testresources.OptimisingTestSuite()
    result.addTests(generate_scenarios(suite))
    for pkgname in packages:
        pkg = __import__('l_mirror.tests.' + pkgname, globals(),
            locals(), ['test_suite'])
        result.addTests(generate_scenarios(pkg.test_suite()))
    return result



#
# LMirror is Copyright (C) 2010 Robert Collins <robertc@robertcollins.net>
# 
# LMirror is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
# 
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE.  See the GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <http://www.gnu.org/licenses/>.
# 
# In the LMirror source tree the file COPYING.txt contains the GNU General Public
# License version 3.
# 

"""Tell receivers a change is being made to a mirror set."""

from bzrlib import urlutils

from l_mirror.arguments import path
from l_mirror.commands import Command
from l_mirror import mirrorset

class start_change(Command):
    """Start a new change to a mirror set.
    
    This tells receivers that they may not be able to read the file content
    they want as it is being changed - so when a mismatch is detected they
    can handle this differently than a fatal error.

    Typically you would run this before any cron job or the like which may
    either delete or change paths already on disk.
    
    Takes the mirror set to start a change on.
    """

    args = [path.PathArgument('mirror_set', min=1, max=1)]

    def run(self):
        transport = self.ui.arguments['mirror_set'][0]
        base = transport.clone('..')
        name = base.relpath(transport.base)
        mirror = mirrorset.MirrorSet(base, name, self.ui)
        mirror.start_change()
        return 0



#
# LMirror is Copyright (C) 2010 Robert Collins <robertc@robertcollins.net>
# 
# LMirror is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
# 
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE.  See the GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <http://www.gnu.org/licenses/>.
# 
# In the LMirror source tree the file COPYING.txt contains the GNU General Public
# License version 3.
# 

"""Am object based UI for l_mirror."""

from cStringIO import StringIO
import optparse

from l_mirror import ui

class ProcessModel(object):
    """A subprocess.Popen test double."""

    def __init__(self, ui):
        self.ui = ui
        self.returncode = 0

    def communicate(self):
        self.ui.outputs.append(('communicate',))
        return '', ''


class UI(ui.AbstractUI):
    """A object based UI.
    
    This is useful for reusing the Command objects that provide a simplified
    interaction model with the domain logic from python. It is used for
    testing l_mirror commands.
    """

    def __init__(self, input_streams=None, options=(), args=(), log_level=4):
        """Create a model UI.

        :param input_streams: A list of stream name, bytes stream tuples to be
            used as the available input streams for this ui.
        :param options: Options to explicitly set values for. A list of
            (option, value) items.
        :param args: The argument values to give the UI.
        :param log_level: Log level to filter on by default.
        """
        self.input_streams = {}
        if input_streams:
            for stream_type, stream_bytes in input_streams:
                self.input_streams.setdefault(stream_type, []).append(
                    stream_bytes)
        self.here = 'memory:'
        self.unparsed_opts = options
        self.outputs = []
        # Could take parsed args, but for now this is easier.
        self.unparsed_args = args
        self.log_level = log_level

    def _check_cmd(self):
        options = list(self.unparsed_opts)
        self.options = optparse.Values()
        seen_options = set()
        for option, value in options:
            setattr(self.options, option, value)
            seen_options.add(option)
        if not 'quiet' in seen_options:
            setattr(self.options, 'quiet', False)
        for option in self.cmd.options:
            if not option.dest in seen_options:
                setattr(self.options, option.dest, option.default)
        args = list(self.unparsed_args)
        parsed_args = {}
        failed = False
        for arg in self.cmd.args:
            try:
                parsed_args[arg.name] = arg.parse(args)
            except ValueError:
                failed = True
                break
        self.arguments = parsed_args
        return args == [] and not failed

    def _iter_streams(self, stream_type):
        streams = self.input_streams.pop(stream_type, [])
        for stream_bytes in streams:
            yield StringIO(stream_bytes)

    def output_error(self, error_tuple):
        self.outputs.append(('error', error_tuple))

    def output_log(self, level, section, line):
        if level > self.log_level:
            self.outputs.append(('log', level, section, line))

    def output_rest(self, rest_string):
        self.outputs.append(('rest', rest_string))

    def output_stream(self, stream):
        self.outputs.append(('stream', stream.read()))

    def output_table(self, table):
        self.outputs.append(('table', table))

    def output_values(self, values):
        self.outputs.append(('values', values))

    def subprocess_Popen(self, *args, **kwargs):
        # Really not an output - outputs should be renamed to events.
        self.outputs.append(('popen', args, kwargs))
        return ProcessModel(self)



#
# LMirror is Copyright (C) 2010 Robert Collins <robertc@robertcollins.net>
# 
# LMirror is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
# 
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE.  See the GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <http://www.gnu.org/licenses/>.
# 
# In the LMirror source tree the file COPYING.txt contains the GNU General Public
# License version 3.
# 

"""A TestResource that provides a temporary python package."""

import os.path
import shutil
import tempfile

from testresources import TestResource

class TempDirResource(TestResource):
    """A temporary directory resource.  

    This resource is never considered dirty.
    """

    def make(self, dependency_resources):
        return tempfile.mkdtemp()

    def clean(self, resource):
        shutil.rmtree(resource, ignore_errors=True)


class StubPackage(object):
    """A temporary package for tests.
    
    :ivar base: The directory containing the package dir.
    """


class StubPackageResource(TestResource):
    
    def __init__(self, packagename, modulelist, init=True):
        super(StubPackageResource, self).__init__()
        self.packagename = packagename
        self.modulelist = modulelist
        self.init = init
        self.resources = [('base', TempDirResource())]

    def make(self, dependency_resources):
        result = StubPackage()
        base = dependency_resources['base']
        root = os.path.join(base, self.packagename)
        os.mkdir(root)
        init_seen = not self.init
        for modulename, contents in self.modulelist:
            stream = file(os.path.join(root, modulename), 'wb')
            try:
                stream.write(contents)
            finally:
                stream.close()
            if modulename == '__init__.py':
                init_seen = True
        if not init_seen:
            file(os.path.join(root, '__init__.py'), 'wb').close()
        return result



#
# LMirror is Copyright (C) 2010 Robert Collins <robertc@robertcollins.net>
# 
# LMirror is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
# 
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE.  See the GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <http://www.gnu.org/licenses/>.
# 
# In the LMirror source tree the file COPYING.txt contains the GNU General Public
# License version 3.
# 

"""Tests for the arguments package."""

from l_mirror import arguments
from l_mirror.tests import ResourcedTestCase


class TestAbstractArgument(ResourcedTestCase):

    def test_init_base(self):
        arg = arguments.AbstractArgument('name')
        self.assertEqual('name', arg.name)
        self.assertEqual('name', arg.summary())

    def test_init_optional(self):
        arg = arguments.AbstractArgument('name', min=0)
        self.assertEqual(0, arg.minimum_count)
        self.assertEqual('name?', arg.summary())

    def test_init_repeating(self):
        arg = arguments.AbstractArgument('name', max=None)
        self.assertEqual(None, arg.maximum_count)
        self.assertEqual('name+', arg.summary())

    def test_init_optional_repeating(self):
        arg = arguments.AbstractArgument('name', min=0, max=None)
        self.assertEqual(None, arg.maximum_count)
        self.assertEqual('name*', arg.summary())

    def test_init_arbitrary(self):
        arg = arguments.AbstractArgument('name', max=2)
        self.assertEqual('name{1,2}', arg.summary())

    def test_init_arbitrary_infinite(self):
        arg = arguments.AbstractArgument('name', min=2, max=None)
        self.assertEqual('name{2,}', arg.summary())

    def test_parsing_calls__parse_one(self):
        calls = []
        class AnArgument(arguments.AbstractArgument):
            def _parse_one(self, arg):
                calls.append(arg)
                return ('1', arg)
        argument = AnArgument('foo', max=2)
        args = ['thing', 'other', 'stranger']
        # results are returned
        self.assertEqual([('1', 'thing'), ('1', 'other')],
            argument.parse(args))
        # used args are removed
        self.assertEqual(['stranger'], args)
        # parse function was used
        self.assertEqual(['thing', 'other'], calls)

    def test_parsing_unlimited(self):
        class AnArgument(arguments.AbstractArgument):
            def _parse_one(self, arg):
                return arg
        argument = AnArgument('foo', max=None)
        args = ['thing', 'other']
        # results are returned
        self.assertEqual(['thing', 'other'], argument.parse(args))
        # used args are removed
        self.assertEqual([], args)

    def test_parsing_too_few(self):
        class AnArgument(arguments.AbstractArgument):
            def _parse_one(self, arg):
                return arg
        argument = AnArgument('foo')
        self.assertRaises(ValueError, argument.parse, [])


# No interface tests for now, because the interface we expect is really just
# _parse_one; however if bugs or issues show up... then we should add them.



#
# LMirror is Copyright (C) 2010 Robert Collins <robertc@robertcollins.net>
# 
# LMirror is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
# 
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE.  See the GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <http://www.gnu.org/licenses/>.
# 
# In the LMirror source tree the file COPYING.txt contains the GNU General Public
# License version 3.
# 

"""Tests for the path argument type."""

from bzrlib import transport
from l_mirror.arguments import path
from l_mirror.tests import ResourcedTestCase


class TestArgument(ResourcedTestCase):

    def test_parses_as_transport(self):
        arg = path.PathArgument('name')
        result = arg.parse(['load'])
        self.assertEqual([transport.get_transport('load').base], [result[0].base])
        self.assertEqual(1, len(result))

    def test_NotLocal_raises_ValueError(self):
        arg = path.PathArgument('name')
        self.assertRaises(ValueError, arg.parse, ['http://foo/'])



#
# LMirror is Copyright (C) 2010 Robert Collins <robertc@robertcollins.net>
# 
# LMirror is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
# 
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE.  See the GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <http://www.gnu.org/licenses/>.
# 
# In the LMirror source tree the file COPYING.txt contains the GNU General Public
# License version 3.
# 

"""Tests for setup.py."""

import doctest
import os.path
import subprocess

from testtools import (
    TestCase,
    )
from testtools.matchers import (
    DocTestMatches,
    )

class TestCanSetup(TestCase):

    def test_bdist(self):
        # Single smoke test to make sure we can build a package.
        path = os.path.join(os.path.dirname(__file__), '..', '..', 'setup.py')
        proc = subprocess.Popen([path, 'bdist'], stdin=subprocess.PIPE,
            stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
        output, _ = proc.communicate()
        self.assertEqual(0, proc.returncode)
        self.assertThat(output, DocTestMatches("""...
...bin/lmirror ...
""", doctest.ELLIPSIS))



#
# LMirror is Copyright (C) 2010 Robert Collins <robertc@robertcollins.net>
# 
# LMirror is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
# 
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE.  See the GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <http://www.gnu.org/licenses/>.
# 
# In the LMirror source tree the file COPYING.txt contains the GNU General Public
# License version 3.
# 

"""'Arguments' for lmirror.

This is a small typed arguments concept - which is perhaps obsoleted by
argparse in Python 2.7, but for l_mirror is an extension used with
optparse.

The code in this module contains the AbstractArgument base class. Individual
argument types are present in e.g. l_mirror.arguments.string.

See l_mirror.commands.Command for usage of Arguments.

Plugins and extensions wanting to add argument types should either define them
internally or install into l_mirror.arguments as somename (perhaps by
extending the l_mirror.arguments __path__ to include a directory
containing their argument types - no __init__ is needed in that directory.)
"""


class AbstractArgument(object):
    """A argument that a command may need.

    Arguments can be converted into a summary for describing the UI to users,
    and provide validator/parsers for the arguments.

    :ivar: The name of the argument. This is used for retrieving the argument
        from UI objects, and for generating the summary.
    """

    def __init__(self, name, min=1, max=1):
        """Create an AbstractArgument.

        While conceptually a separate SequenceArgument could be used, all
        arguments support sequencing to avoid unnecessary boilerplate in user
        code.

        :param name: The name for the argument.
        :param min: The minimum number of occurences permitted.
        :param max: The maximum number of occurences permitted. None for
            unlimited.
        """
        self.name = name
        self.minimum_count = min
        self.maximum_count = max

    def summary(self):
        """Get a regex-like summary of this argument."""
        result = self.name
        if (self.minimum_count == self.maximum_count and
            self.minimum_count == 1):
                return result
        minmax = (self.minimum_count, self.maximum_count)
        if minmax == (0, 1):
            return result + '?'
        if minmax == (1, None):
            return result + '+'
        if minmax == (0, None):
            return result + '*'
        if minmax[1] == None:
            minmax = (minmax[0], '')
        return result + '{%s,%s}' % minmax

    def parse(self, argv):
        """Evaluate arguments in argv.

        Used arguments are removed from argv.

        :param argv: The arguments to parse.
        :return: The parsed results as a list.
        """
        count = 0
        result = []
        while len(argv) > count and (
            count < self.maximum_count or self.maximum_count is None):
            arg = argv[count]
            count += 1
            result.append(self._parse_one(arg))
        if count < self.minimum_count:
            raise ValueError('not enough arguments present in %s' % argv)
        del argv[:count]
        return result

    def _parse_one(self, arg):
        """Parse a single argument.
        
        :param arg: An arg from an argv.
        :result: The parsed argument.
        :raises ValueError: If the arg cannot be parsed/validated.
        """
        raise NotImplementedError(self._parse_one)



#
# LMirror is Copyright (C) 2010 Robert Collins <robertc@robertcollins.net>
# 
# LMirror is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
# 
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE.  See the GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <http://www.gnu.org/licenses/>.
# 
# In the LMirror source tree the file COPYING.txt contains the GNU General Public
# License version 3.
# 

"""Tests for UI support logic and the UI contract."""

import doctest
from cStringIO import StringIO
import sys

from testtools.matchers import DocTestMatches

from l_mirror import arguments
import l_mirror.arguments.command
import l_mirror.arguments.string
from l_mirror import commands
from l_mirror.ui import cli
from l_mirror.tests import ResourcedTestCase
from l_mirror.tests.logging_resource import LoggingResourceManager
from l_mirror.tests.monkeypatch import monkeypatch
from l_mirror.tests.stubhome import StubHomeResource


class TestCLIUI(ResourcedTestCase):

    resources = [('logging', LoggingResourceManager())]

    def get_test_ui_and_cmd(self):
        stdout = StringIO()
        stdin = StringIO()
        stderr = StringIO()
        ui = cli.UI([], stdin, stdout, stderr, no_logfile=True)
        cmd = commands.Command(ui)
        ui.set_command(cmd)
        return ui, cmd

    def test_construct(self):
        stdout = StringIO()
        stdin = StringIO()
        stderr = StringIO()
        ui = cli.UI([], stdin, stdout, stderr, no_logfile=True)

    def test_stream_comes_from_stdin(self):
        stdout = StringIO()
        stdin = StringIO('foo\n')
        stderr = StringIO()
        ui = cli.UI([], stdin, stdout, stderr, no_logfile=True)
        cmd = commands.Command(ui)
        cmd.input_streams = ['content']
        ui.set_command(cmd)
        results = []
        for stream in ui.iter_streams('content'):
            results.append(stream.read())
        self.assertEqual(['foo\n'], results)

    def test_dash_d_sets_here_option(self):
        stdout = StringIO()
        stdin = StringIO('foo\n')
        stderr = StringIO()
        ui = cli.UI(['-d', '/nowhere/'], stdin, stdout, stderr, no_logfile=True)
        cmd = commands.Command(ui)
        ui.set_command(cmd)
        self.assertEqual('/nowhere/', ui.here)

    def test_global_option_dash_v_changes_log_level(self):
        stdout = StringIO()
        stdin = StringIO()
        stderr = StringIO()
        ui = cli.UI(['-v'], stdin, stdout, stderr, no_logfile=True)
        cmd = commands.Command(ui)
        ui.set_command(cmd)
        self.assertEqual(4, ui._c_log.level)

    def test_outputs_error_string(self):
        try:
            raise Exception('fooo')
        except Exception:
            err_tuple = sys.exc_info()
        expected = str(err_tuple[1]) + '\n'
        stdout = StringIO()
        stdin = StringIO()
        stderr = StringIO()
        ui = cli.UI([], stdin, stdout, stderr, no_logfile=True)
        ui.output_error(err_tuple)
        self.assertThat(stderr.getvalue(), DocTestMatches(expected))

    def test_outputs_log_to_stdout(self):
        ui, cmd = self.get_test_ui_and_cmd()
        ui.output_log(5, 'my.self', 'line')
        self.assertThat(ui._stdout.getvalue(),
            DocTestMatches("""line\n""", doctest.ELLIPSIS))

    def test_outputs_rest_to_stdout(self):
        ui, cmd = self.get_test_ui_and_cmd()
        ui.output_rest('topic\n=====\n')
        self.assertEqual('topic\n=====\n', ui._stdout.getvalue())

    def test_outputs_stream_to_stdout(self):
        ui, cmd = self.get_test_ui_and_cmd()
        stream = StringIO("Foo \n bar")
        ui.output_stream(stream)
        self.assertEqual("Foo \n bar", ui._stdout.getvalue())

    def test_outputs_tables_to_stdout(self):
        ui, cmd = self.get_test_ui_and_cmd()
        ui.output_table([('foo', 1), ('b', 'quux')])
        self.assertEqual('foo  1\n---  ----\nb    quux\n',
            ui._stdout.getvalue())

    def test_outputs_values_to_stdout(self):
        ui, cmd = self.get_test_ui_and_cmd()
        ui.output_values([('foo', 1), ('bar', 'quux')])
        self.assertEqual('foo: 1 bar: quux\n', ui._stdout.getvalue())

    def test_parse_error_goes_to_stderr(self):
        stdout = StringIO()
        stdin = StringIO()
        stderr = StringIO()
        ui = cli.UI(['one'], stdin, stdout, stderr, no_logfile=True)
        cmd = commands.Command(ui)
        cmd.args = [arguments.command.CommandArgument('foo')]
        ui.set_command(cmd)
        self.assertEqual("Could not find command 'one'.\n", stderr.getvalue())

    def test_parse_excess_goes_to_stderr(self):
        stdout = StringIO()
        stdin = StringIO()
        stderr = StringIO()
        ui = cli.UI(['one'], stdin, stdout, stderr, no_logfile=True)
        cmd = commands.Command(ui)
        ui.set_command(cmd)
        self.assertEqual("Unexpected arguments: ['one']\n", stderr.getvalue())

    def test_parse_after_double_dash_are_arguments(self):
        stdout = StringIO()
        stdin = StringIO()
        stderr = StringIO()
        ui = cli.UI(['one', '--', '--two', 'three'], stdin, stdout, stderr,
            no_logfile=True)
        cmd = commands.Command(ui)
        cmd.args = [arguments.string.StringArgument('args', max=None)]
        ui.set_command(cmd)
        self.assertEqual({'args':['one', '--two', 'three']}, ui.arguments)


class TestRunArgv(ResourcedTestCase):

    def stub__find_command(self, cmd_run):
        self.calls = []
        self.addCleanup(monkeypatch('l_mirror.commands._find_command',
            self._find_command))
        self.cmd_run = cmd_run

    def _find_command(self, cmd_name):
        self.calls.append(cmd_name)
        real_run = self.cmd_run
        class SampleCommand(commands.Command):
            """A command that is used for testing."""
            def execute(self):
                return real_run(self)
        return SampleCommand

    def test_looks_up_cmd(self):
        self.stub__find_command(lambda x:0)
        cli.run_argv(['lmirror', 'foo'], 'in', 'out', 'err')
        self.assertEqual(['foo'], self.calls)

    def test_looks_up_cmd_skips_options(self):
        self.stub__find_command(lambda x:0)
        cli.run_argv(['lmirror', '--version', 'foo'], 'in', 'out', 'err')
        self.assertEqual(['foo'], self.calls)

    def test_no_cmd_issues_help(self):
        self.stub__find_command(lambda x:0)
        cli.run_argv(['lmirror', '--version'], 'in', 'out', 'err')
        self.assertEqual(['help'], self.calls)

    def capture_ui(self, cmd):
        self.ui = cmd.ui
        return 0

    def test_runs_cmd_with_CLI_UI(self):
        self.stub__find_command(self.capture_ui)
        cli.run_argv(['lmirror', '--version', 'foo'], 'in', 'out', 'err')
        self.assertEqual(['foo'], self.calls)
        self.assertIsInstance(self.ui, cli.UI)

    def test_returns_0_when_None_returned_from_execute(self):
        self.stub__find_command(lambda x:None)
        self.assertEqual(0, cli.run_argv(['lmirror', 'foo'], 'in', 'out',
            'err'))

    def test_returns_execute_result(self):
        self.stub__find_command(lambda x:1)
        self.assertEqual(1, cli.run_argv(['lmirror', 'foo'], 'in', 'out',
            'err'))


class TestCLILogFile(ResourcedTestCase):

    resources = [('logging', LoggingResourceManager()), 
        ('homedir', StubHomeResource())]

    def get_test_ui_and_cmd(self):
        stdout = StringIO()
        stdin = StringIO()
        stderr = StringIO()
        ui = cli.UI([], stdin, stdout, stderr)
        cmd = commands.Command(ui)
        ui.set_command(cmd)
        return ui, cmd

    def test_outputs_log_to_homedir(self):
        ui, cmd = self.get_test_ui_and_cmd()
        ui.output_log(8, 'my.self', 'line')
        ui.output_log(2, 'my.self', 'not logged line')
        content = file(self.homedir.tempdir + '/.cache/lmirror/log', 'rb').read()
        self.assertThat(content,
            DocTestMatches("""...Z: line\n""", doctest.ELLIPSIS))




#
# LMirror is Copyright (C) 2010 Robert Collins <robertc@robertcollins.net>
# 
# LMirror is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
# 
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE.  See the GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <http://www.gnu.org/licenses/>.
# 
# In the LMirror source tree the file COPYING.txt contains the GNU General Public
# License version 3.
# 

"""Tests for the journals module."""

from doctest import ELLIPSIS
import time

from bzrlib.transport import get_transport
from bzrlib.transport.memory import MemoryServer

from testtools.matchers import DocTestMatches

from l_mirror import journals
from l_mirror.ui.model import UI
from l_mirror.tests import ResourcedTestCase


class TestCombiner(ResourcedTestCase):

    def test_init(self):
        combiner = journals.Combiner()

    def test_add_journal_no_conflict(self):
        combiner = journals.Combiner()
        j1 = journals.Journal()
        j1.add('abc', 'new', ('file', '12039d6dd9a7e27622301e935b6eefc78846802e', 11))
        combiner.add(j1)

    def test_add_delete_no_conflict(self):
        combiner = journals.Combiner()
        j1 = journals.Journal()
        j1.add('abc', 'del', ('file', '12039d6dd9a7e27622301e935b6eefc78846802e', 11))
        combiner.add(j1)

    def test_add_replace_no_path_no_conflict(self):
        combiner = journals.Combiner()
        j1 = journals.Journal()
        j1.add('abc', 'replace', (
            ('file', '12039d6dd9a7e27622301e935b6eefc78846802e', 11),
            ('file', 'abcdef0123412312341234123412341234123412', 34)))
        combiner.add(j1)

    def test_add_new_new_conflicts(self):
        combiner = journals.Combiner()
        j1 = journals.Journal()
        j1.add('abc', 'new', ('file', '12039d6dd9a7e27622301e935b6eefc78846802e', 11))
        combiner.add(j1)
        self.assertRaises(ValueError, combiner.add, j1)

    def test_add_new_delete_mismatch_path_conflicts(self):
        combiner = journals.Combiner()
        j1 = journals.Journal()
        j1.add('abc', 'new',
            ('file', '12039d6dd9a7e27622301e935b6eefc78846802e', 11))
        combiner.add(j1)
        j2 = journals.Journal()
        j2.add('abc', 'del',
            ('file', 'abc39d6dd9a7e27622301e935b6eefc78846802e', 11))
        self.assertRaises(ValueError, combiner.add, j2)

    def test_add_new_delete_no_conflicts(self):
        combiner = journals.Combiner()
        j1 = journals.Journal()
        j1.add('abc', 'new',
            ('file', '12039d6dd9a7e27622301e935b6eefc78846802e', 11))
        combiner.add(j1)
        j2 = journals.Journal()
        j2.add('abc', 'del',
            ('file', '12039d6dd9a7e27622301e935b6eefc78846802e', 11))
        combiner.add(j2)
        self.assertEqual({}, combiner.journal.paths)

    def test_add_new_replace_mismatch_path_conflicts(self):
        combiner = journals.Combiner()
        j1 = journals.Journal()
        j1.add('abc', 'new',
            ('file', '12039d6dd9a7e27622301e935b6eefc78846802e', 11))
        combiner.add(j1)
        j2 = journals.Journal()
        j2.add('abc', 'replace', (
            ('file', 'abc39d6dd9a7e27622301e935b6eefc78846802e', 11),
            ('file', 'abcdef0123412312341234123412341234123412', 34)))
        self.assertRaises(ValueError, combiner.add, j2)

    def test_add_new_replace_no_conflicts(self):
        combiner = journals.Combiner()
        j1 = journals.Journal()
        j1.add('abc', 'new',
            ('file', '12039d6dd9a7e27622301e935b6eefc78846802e', 11))
        combiner.add(j1)
        j2 = journals.Journal()
        j2.add('abc', 'replace', (
            ('file', '12039d6dd9a7e27622301e935b6eefc78846802e', 11),
            ('file', 'abcdef0123412312341234123412341234123412', 34)))
        combiner.add(j2)
        self.assertEqual({'abc': ('new', 
            ('file', 'abcdef0123412312341234123412341234123412', 34))},
            combiner.journal.paths)

    def test_add_replace_del_no_conflicts(self):
        combiner = journals.Combiner()
        j1 = journals.Journal()
        j1.add('abc', 'replace', (
            ('file', '12039d6dd9a7e27622301e935b6eefc78846802e', 11),
            ('file', 'abcdef0123412312341234123412341234123412', 34)))
        combiner.add(j1)
        j2 = journals.Journal()
        j2.add('abc', 'del',
            ('file', 'abcdef0123412312341234123412341234123412', 34))
        combiner.add(j2)
        self.assertEqual({'abc': ('del', 
            ('file', '12039d6dd9a7e27622301e935b6eefc78846802e', 11))},
            combiner.journal.paths)

    def test_add_replace_del_mismatch_path_conflicts(self):
        combiner = journals.Combiner()
        j1 = journals.Journal()
        j1.add('abc', 'replace', (
            ('file', '12039d6dd9a7e27622301e935b6eefc78846802e', 11),
            ('file', 'abcdef0123412312341234123412341234123412', 34)))
        combiner.add(j1)
        j2 = journals.Journal()
        j2.add('abc', 'del',
            ('file', '12339d6dd9a7e27622301e935b6eefc78846802e', 34))
        self.assertRaises(ValueError, combiner.add, j2)

    def test_add_replace_new_conflicts(self):
        combiner = journals.Combiner()
        j1 = journals.Journal()
        j1.add('abc', 'replace', (
            ('file', '12039d6dd9a7e27622301e935b6eefc78846802e', 11),
            ('file', 'abcdef0123412312341234123412341234123412', 34)))
        combiner.add(j1)
        j2 = journals.Journal()
        j2.add('abc', 'new',
            ('file', 'abcdef0123412312341234123412341234123412', 34))
        self.assertRaises(ValueError, combiner.add, j2)

    def test_add_replace_replace_no_conflicts(self):
        combiner = journals.Combiner()
        j1 = journals.Journal()
        j1.add('abc', 'replace', (
            ('file', '12039d6dd9a7e27622301e935b6eefc78846802e', 11),
            ('file', 'abcdef0123412312341234123412341234123412', 34)))
        combiner.add(j1)
        j2 = journals.Journal()
        j2.add('abc', 'replace', (
            ('file', 'abcdef0123412312341234123412341234123412', 34),
            ('file', '9999999999999999999999999999999999999999', 15)))
        combiner.add(j2)
        self.assertEqual({'abc': ('replace', (
            ('file', '12039d6dd9a7e27622301e935b6eefc78846802e', 11),
            ('file', '9999999999999999999999999999999999999999', 15)))},
            combiner.journal.paths)

    def test_add_replace_replace_mismatch_path_conflicts(self):
        combiner = journals.Combiner()
        j1 = journals.Journal()
        j1.add('abc', 'replace', (
            ('file', '12039d6dd9a7e27622301e935b6eefc78846802e', 11),
            ('file', 'abcdef0123412312341234123412341234123412', 34)))
        combiner.add(j1)
        j2 = journals.Journal()
        j2.add('abc', 'replace', (
            ('file', 'abc39d6dd9a7e27622301e935b6eefc78846802e', 11),
            ('file', 'abcdef0123412312341234123412341234123412', 34)))
        self.assertRaises(ValueError, combiner.add, j2)

    def test_add_del_del_conflicts(self):
        combiner = journals.Combiner()
        j1 = journals.Journal()
        j1.add('abc', 'del', ('file', '12039d6dd9a7e27622301e935b6eefc78846802e', 11))
        combiner.add(j1)
        self.assertRaises(ValueError, combiner.add, j1)

    def test_add_del_replace_conflicts(self):
        combiner = journals.Combiner()
        j1 = journals.Journal()
        j1.add('abc', 'del', ('file', '12039d6dd9a7e27622301e935b6eefc78846802e', 11))
        combiner.add(j1)
        j2 = journals.Journal()
        j2.add('abc', 'replace', (
            ('file', '12039d6dd9a7e27622301e935b6eefc78846802e', 11),
            ('file', '12039d6dd9a7e27622301e935b6eefc78846802e', 11)))
        self.assertRaises(ValueError, combiner.add, j2)

    def test_add_del_new_ok(self):
        combiner = journals.Combiner()
        j1 = journals.Journal()
        j1.add('abc', 'del', ('file', '12039d6dd9a7e27622301e935b6eefc78846802e', 11))
        combiner.add(j1)
        j2 = journals.Journal()
        j2.add('abc', 'new', ('file', '00039d6dd9a7e27622301e935b6eefc78846802e', 34))
        combiner.add(j2)

    def test_as_tree_empty(self):
        combiner = journals.Combiner()
        self.assertEqual({}, combiner.as_tree())

    def test_as_tree_demo(self):
        combiner = journals.Combiner()
        file1 = ('file', '12039d6dd9a7e27622301e935b6eefc78846802e', 11)
        file2 = ('file', 'abcdef0123412312341234123412341234123412', 34)
        link1 = ('symlink', 'foo bar/baz')
        dir1 = ('dir',)
        expected = {'foo': file1,
                    'bar': {
                        'gam': file2,
                        'quux': link1,
                    },
                    'gam': {
                    }
            }
        j1 = journals.Journal()
        j1.add('foo', 'new', file1)
        j1.add('gam', 'new', dir1)
        combiner.add(j1)
        j2 = journals.Journal()
        j2.add('bar', 'new', dir1)
        j2.add('bar/gam', 'new', file1)
        combiner.add(j2)
        j3 = journals.Journal()
        j3.add('bar/gam', 'replace', (file1, file2))
        j3.add('bar/quux', 'new', link1)
        combiner.add(j3)
        self.assertEqual(expected, combiner.as_tree())

    def test_as_tree_replace_fail(self):
        combiner = journals.Combiner()
        j1 = journals.Journal()
        j1.add('foo', 'replace', (('dir',), ('dir',)))
        combiner.add(j1)
        self.assertRaises(ValueError, combiner.as_tree)

    def test_as_tree_del_fail(self):
        combiner = journals.Combiner()
        j1 = journals.Journal()
        j1.add('foo', 'del', ('dir',))
        combiner.add(j1)
        self.assertRaises(ValueError, combiner.as_tree)

    def test_missing_parent_fail(self):
        combiner = journals.Combiner()
        j1 = journals.Journal()
        j1.add('foo/bar', 'new', ('dir',))
        combiner.add(j1)
        self.assertRaises(ValueError, combiner.as_tree)


class TestJournal(ResourcedTestCase):

    def test_add_new_ok(self):
        j1 = journals.Journal()
        j1.add('abc', 'new', ('file',
            '12039d6dd9a7e27622301e935b6eefc78846802e', 11))
        self.assertTrue('abc' in j1.paths)

    def test_add_dup_path_error(self):
        j1 = journals.Journal()
        j1.add('abc', 'new', ('file',
            '12039d6dd9a7e27622301e935b6eefc78846802e', 11))
        self.assertRaises(ValueError, j1.add, 'abc', 'new', ('file',
            '12039d6dd9a7e27622301e935b6eefc78846802e', 11))

    def test_del_new_ok(self):
        j1 = journals.Journal()
        j1.add('abc', 'del', ('file',
            '12039d6dd9a7e27622301e935b6eefc78846802e', 11))
        self.assertTrue('abc' in j1.paths)

    def test_del_dup_path_error(self):
        j1 = journals.Journal()
        j1.add('abc', 'del', ('file',
            '12039d6dd9a7e27622301e935b6eefc78846802e', 11))
        self.assertRaises(ValueError, j1.add, 'abc', 'del', ('file',
            '12039d6dd9a7e27622301e935b6eefc78846802e', 11))

    def test_replace_new_ok(self):
        j1 = journals.Journal()
        j1.add('abc', 'replace', (('file',
            '12039d6dd9a7e27622301e935b6eefc78846802e', 11), ('file',
            'e935b6eefc78846802e12039d6dd9a7e27622301', 12)))
        self.assertTrue('abc' in j1.paths)

    def test_replace_dup_path_error(self):
        j1 = journals.Journal()
        j1.add('abc', 'replace', (('file',
            '12039d6dd9a7e27622301e935b6eefc78846802e', 11), ('file',
            'e935b6eefc78846802e12039d6dd9a7e27622301', 12)))
        self.assertRaises(ValueError, j1.add, 'abc', 'replace', (('file',
            '12039d6dd9a7e27622301e935b6eefc78846802e', 11), ('file',
            'e935b6eefc78846802e12039d6dd9a7e27622301', 12)))

    def test_replace_single_details_errors(self):
        j1 = journals.Journal()
        j1.add('abc', 'replace', (
            ('file', '12039d6dd9a7e27622301e935b6eefc78846802e', 11),
            ('file', 'abcdef1231231231231231231241231231231241', 12)))

    def test_as_bytes(self):
        # as bytes can serialise all types, and is sorted by path.
        j1 = journals.Journal()
        j1.add('abc', 'new', ('file',
            '12039d6dd9a7e27622301e935b6eefc78846802e', 11))
        j1.add('abc/def', 'del', ('dir',))
        j1.add('1234', 'replace', (('symlink', 'foo bar/baz'), ('file', 'e935b6eefc78846802e12039d6dd9a7e27622301', 0)))
        self.assertEqual("""l-mirror-journal-1
1234\0replace\0symlink\0foo bar/baz\0file\0e935b6eefc78846802e12039d6dd9a7e27622301\0000\x00abc\0new\0file\00012039d6dd9a7e27622301e935b6eefc78846802e\00011\0abc/def\0del\0dir""", j1.as_bytes())


class TestTransportReplay(ResourcedTestCase):

    def setup_memory(self):
        """Create a memory url server and return its url."""
        # XXX: integrate with ui.here better.
        self.transport_factory = MemoryServer()
        self.transport_factory.start_server()
        self.addCleanup(self.transport_factory.stop_server)
        return self.transport_factory.get_url()

    def test_orders_new_replace_delete(self):
        # new-replace-delete is a sane default order.
        basedir = get_transport('trace+' + self.setup_memory()).clone('path')
        basedir.create_prefix()
        basedir.put_bytes('abc', 'def')
        basedir.put_bytes('bye', 'by')
        sourcedir = basedir.clone('../source')
        sourcedir.create_prefix()
        sourcedir.put_bytes('abc', '123412341234')
        sourcedir.put_bytes('new', '12341234')
        j1 = journals.Journal()
        j1.add('abc', 'replace', (
            ('file', '12039d6dd9a7e27622301e935b6eefc78846802e', 3),
            ('file', '5a78babbb162531b3a16c55310a4e7228d68f2e9', 12)))
        j1.add('bye', 'del', ('file', 'd', 2))
        j1.add('new', 'new', ('file', 'c129b324aee662b04eccf68babba85851346dff9', 8))
        del basedir._activity[:]
        ui = UI()
        replay = journals.TransportReplay(j1, sourcedir, basedir, ui)
        replay.replay()
        self.assertEqual([('get', 'new'), ('rename', 'new.lmirrortemp', 'new'),
            ('get', 'abc'), ('get', 'abc'), ('delete', 'abc'), ('rename',
            'abc.lmirrortemp', 'abc'), ('delete', 'bye')],
            basedir._activity)


class TestParser(ResourcedTestCase):

    def test_parse_demo(self):
        journal = journals.parse("""l-mirror-journal-1
1234\0replace\0symlink\0foo bar/baz\0file\0e935b6eefc78846802e12039d6dd9a7e27622301\0000\x00abc\0new\0file\00012039d6dd9a7e27622301e935b6eefc78846802e\00011\0abc/def\0del\0dir""")
        expected = journals.Journal()
        expected.add('abc', 'new', ('file',
            '12039d6dd9a7e27622301e935b6eefc78846802e', 11))
        expected.add('abc/def', 'del', ('dir',))
        expected.add('1234', 'replace', (('symlink', 'foo bar/baz'), ('file', 'e935b6eefc78846802e12039d6dd9a7e27622301', 0)))
        self.assertEqual(expected.paths, journal.paths)

    def test_parse_empty(self):
        journal = journals.parse('l-mirror-journal-1\n')
        self.assertEqual({}, journal.paths)

    def test_parse_wrong_header(self):
        self.assertRaises(ValueError, journals.parse, 'l-mirror-journal-1')
        self.assertRaises(ValueError, journals.parse, 'l-mirror-journal-2\n')


class TestDiskUpdater(ResourcedTestCase):

    def get_test_ui(self):
        ui = UI()
        return ui

    def setup_memory(self):
        """Create a memory url server and return its url."""
        # XXX: integrate with ui.here better.
        self.transport_factory = MemoryServer()
        self.transport_factory.start_server()
        self.addCleanup(self.transport_factory.stop_server)
        return self.transport_factory.get_url()

    def test_update_demo(self):
        ui = self.get_test_ui()
        now = time.time()
        four_seconds = now - 4
        basedir = get_transport(self.setup_memory()).clone('path')
        basedir.create_prefix()
        ui = self.get_test_ui()
        basedir.create_prefix()
        basedir.mkdir('dir1')
        basedir.mkdir('dir2')
        basedir.put_bytes('abc', '1234567890\n')
        basedir.put_bytes('dir1/def', 'abcdef')
        last_timestamp = 0 # get everything
        updater = journals.DiskUpdater({}, basedir, 'name', last_timestamp, ui)
        journal = updater.finished()
        expected = {
            'dir2': ('new', ('dir',)),
            'dir1': ('new', ('dir',)),
            'abc': ('new', ('file', '12039d6dd9a7e27622301e935b6eefc78846802e', 11)),
            'dir1/def': ('new', ('file', '1f8ac10f23c5b5bc1167bda84b833e5c057a77d2', 6))
            }
        self.assertEqual(expected, journal.paths)

    def test_skips_other_sets(self):
        ui = self.get_test_ui()
        now = time.time()
        four_seconds = now - 4
        basedir = get_transport(self.setup_memory()).clone('path')
        basedir.create_prefix()
        ui = self.get_test_ui()
        basedir.create_prefix()
        basedir.mkdir('.lmirror')
        basedir.mkdir('.lmirror/sets')
        basedir.mkdir('.lmirror/sets/name')
        basedir.mkdir('.lmirror/sets/othername')
        basedir.put_bytes('.lmirror/sets/name/abc', '1234567890\n')
        basedir.put_bytes('.lmirror/sets/othername/abc', '1234567890\n')
        last_timestamp = 0 # get everything
        updater = journals.DiskUpdater({}, basedir, 'name', last_timestamp, ui)
        journal = updater.finished()
        expected = {
            '.lmirror': ('new', ('dir',)),
            '.lmirror/sets': ('new', ('dir',)),
            '.lmirror/sets/name': ('new', ('dir',)),
            '.lmirror/sets/name/abc': ('new', ('file', '12039d6dd9a7e27622301e935b6eefc78846802e', 11)),
            }
        self.assertEqual(expected, journal.paths)



#
# LMirror is Copyright (C) 2010 Robert Collins <robertc@robertcollins.net>
# 
# LMirror is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
# 
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE.  See the GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <http://www.gnu.org/licenses/>.
# 
# In the LMirror source tree the file COPYING.txt contains the GNU General Public
# License version 3.
# 

"""Tests for the start-change command."""

from doctest import ELLIPSIS

from bzrlib.transport import get_transport
from bzrlib.transport.memory import MemoryServer

from testtools.matchers import DocTestMatches

from l_mirror.commands import start_change
from l_mirror import mirrorset
from l_mirror.ui.model import UI
from l_mirror.tests import ResourcedTestCase
from l_mirror.tests.matchers import MatchesException


class TestCommandStartChange(ResourcedTestCase):

    def get_test_ui_and_cmd(self, args):
        ui = UI(args=args)
        cmd = start_change.start_change(ui)
        ui.set_command(cmd)
        return ui, cmd

    def setup_memory(self):
        """Create a memory url server and return its url."""
        # XXX: integrate with ui.here better.
        self.transport_factory = MemoryServer()
        self.transport_factory.start_server()
        self.addCleanup(self.transport_factory.stop_server)
        return self.transport_factory.get_url()

    def test_updating_errors(self):
        base = self.setup_memory()
        t = get_transport(base)
        t = t.clone('path')
        t.create_prefix()
        root = t.base + 'myname'
        ui, cmd = self.get_test_ui_and_cmd((root,))
        mirror = mirrorset.initialise(t, 'myname', t, UI())
        self.assertEqual(3, cmd.execute())
        self.assertEqual(1, len(ui.outputs))
        self.assertEqual('error', ui.outputs[0][0])
        self.assertThat(ui.outputs[0][1],
            MatchesException(ValueError('Changeset already open')))

    def test_not_updating_starts(self):
        base = self.setup_memory()
        root = base + 'path/myname'
        t = get_transport(base)
        t = t.clone('path')
        t.create_prefix()
        ui, cmd = self.get_test_ui_and_cmd((root,))
        mirror = mirrorset.initialise(t, 'myname', t, ui)
        mirror.finish_change()
        self.assertEqual(0, cmd.execute())
        t = t.clone('.lmirror/metadata/myname')
        self.assertThat(t.get_bytes('metadata.conf'), DocTestMatches("""[metadata]
basis = 0
latest = 1
timestamp = ...
updating = True

""", ELLIPSIS))



#
# LMirror is Copyright (C) 2010 Robert Collins <robertc@robertcollins.net>
# 
# LMirror is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
# 
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE.  See the GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <http://www.gnu.org/licenses/>.
# 
# In the LMirror source tree the file COPYING.txt contains the GNU General Public
# License version 3.
# 

"""A command line UI for l_mirror."""

__all__ = ['UI', 'run_argv']

import logging
from optparse import OptionParser
import os
import sys

from l_mirror import commands, ui
from l_mirror import logging_support


class UI(ui.AbstractUI):
    """A command line user interface."""

    def __init__(self, argv, stdin, stdout, stderr, no_logfile=False):
        """Create a command line UI.

        This UI logs to ~/.cache/lmirror/log for level 3 and up by default,
        and level 5 and up to the console.

        :param argv: Arguments from the process invocation.
        :param stdin: The stream for stdin.
        :param stdout: The stream for stdout.
        :param stderr: The stream for stderr.
        ;param no_logfile: Disable default ~/.cache/lmirror/log logfile.
        """
        self._argv = argv
        self._stdin = stdin
        self._stdout = stdout
        self._stderr = stderr
        if no_logfile:
            args = [None]
        else:
            args = []
        self._c_log, self._f_log, self._log_formatter = \
            logging_support.configure_logging(self._stdout, *args)
        self._c_log.setLevel(5)
        if self._f_log is not None:
            self._f_log.setLevel(3) # log more than the UI, but not everything.

    def _iter_streams(self, stream_type):
        yield self._stdin

    def output_error(self, error_tuple):
        self._stderr.write(str(error_tuple[1]) + '\n')
        logging.getLogger().log(3, "Error", exc_info=1)

    def output_log(self, level, section, line):
        logger = logging.getLogger(section)
        logger.log(level, line)

    def output_rest(self, rest_string):
        self._stdout.write(rest_string)
        if not rest_string.endswith('\n'):
            self._stdout.write('\n')

    def output_stream(self, stream):
        contents = stream.read(65536)
        while contents:
            self._stdout.write(contents)
            contents = stream.read(65536)

    def output_table(self, table):
        # stringify
        contents = []
        for row in table:
            new_row = []
            for column in row:
                new_row.append(str(column))
            contents.append(new_row)
        if not contents:
            return
        widths = [0] * len(contents[0])
        for row in contents:
            for idx, column in enumerate(row):
                if widths[idx] < len(column):
                    widths[idx] = len(column)
        # Show a row
        outputs = []
        def show_row(row):
            for idx, column in enumerate(row):
                outputs.append(column)
                if idx == len(row) - 1:
                    outputs.append('\n')
                    return
                # spacers for the next column
                outputs.append(' '*(widths[idx]-len(column)))
                outputs.append('  ')
        show_row(contents[0])
        # title spacer
        for idx, width in enumerate(widths):
            outputs.append('-'*width)
            if idx == len(widths) - 1:
                outputs.append('\n')
                continue
            outputs.append('  ')
        for row in contents[1:]:
            show_row(row)
        self._stdout.write(''.join(outputs))

    def output_values(self, values):
        outputs = []
        for label, value in values:
            outputs.append('%s: %s' % (label, value))
        self._stdout.write('%s\n' % ' '.join(outputs))

    def _check_cmd(self):
        cmd = self.cmd
        parser = OptionParser()
        parser.add_option("-d", "--here", dest="here",
            help="Set the directory or url that a command should run from. "
            "This affects all default path lookups but does not affect paths "
            "supplied to the command.", default=os.getcwd(), type=str)
        parser.add_option("-q", "--quiet", action="store_true", default=False,
            help="Turn off output other than the primary output for a command "
            "and any errors.")
        parser.add_option("-v", "--verbose", action="count", dest="verbosity",
            help="Show more detail about program progression on stdout and in "
            "custom log files.")
        for option in self.cmd.options:
            parser.add_option(option)
        options, args = parser.parse_args(self._argv)
        self.here = options.here
        self.options = options
        if self.options.verbosity:
            self._c_log.setLevel(self._c_log.level - self.options.verbosity)
        orig_args = list(args)
        parsed_args = {}
        failed = False
        for arg in self.cmd.args:
            try:
                parsed_args[arg.name] = arg.parse(args)
            except ValueError:
                exc_info = sys.exc_info()
                failed = True
                self._stderr.write("%s\n" % str(exc_info[1]))
                break
        if not failed:
            self.arguments = parsed_args
            if args != []:
                self._stderr.write("Unexpected arguments: %r\n" % args)
        return not failed and args == []

    def subprocess_Popen(self, *args, **kwargs):
        self.output_log(3, 'l_mirror.ui.cli', "Running subprocess %r %r" % (
            args, kwargs))
        import subprocess
        return subprocess.Popen(*args, **kwargs)


def run_argv(argv, stdin, stdout, stderr):
    """Convenience function to run a command with a CLIUI.

    :param argv: The argv to run the command with.
    :param stdin: The stdin stream for the command.
    :param stdout: The stdout stream for the command.
    :param stderr: The stderr stream for the command.
    :return: An integer exit code for the command.
    """
    cmd_name = None
    cmd_args = argv[1:]
    for arg in argv[1:]:
        if not arg.startswith('-'):
            cmd_name = arg
            break
    if cmd_name is None:
        cmd_name = 'help'
        cmd_args = ['help']
    cmd_args.remove(cmd_name)
    cmdclass = commands._find_command(cmd_name)
    ui = UI(cmd_args, stdin, stdout, stderr)
    cmd = cmdclass(ui)
    result = cmd.execute()
    if not result:
        return 0
    return result



#
# LMirror is Copyright (C) 2010 Robert Collins <robertc@robertcollins.net>
# 
# LMirror is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
# 
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE.  See the GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <http://www.gnu.org/licenses/>.
# 
# In the LMirror source tree the file COPYING.txt contains the GNU General Public
# License version 3.
# 

"""LMirror local gpg support."""

__all__ = ['SimpleGPGStrategy']

import subprocess
import sys

from bzrlib import gpg

class SimpleGPGStrategy(gpg.GPGStrategy):
    """A simple signing strategy that does not need configuration.

    :seealso bzrlib.gpg.GPGStrategy: The base class in bzrlib that needs
        configuration.

    This also does plain signs, rather than clearsigned signatures.
    """

    def _command_line(self):
        return ['gpg', '--detach-sign']


class GPGVStrategy(object):
    """Class for performing GPGV calls.
    
    :ivar keyringpath: The path to the keyring to validate against.
    :ivar ui: A UI object.
    """

    def __init__(self, keyringpath, ui):
        self.keyringpath = keyringpath
        self.ui = ui

    def verify(self, sigdir, signame, content):
        """Check that content is validly signed.

        :param sigdir: Transport that the signature can be found on.
        :param signame: The name of the signature.
        :param content: The content that should be signed.
        """
        sigpath = sigdir.local_abspath(signame)
        # Using the shell gives us solid sigpipe handling, strange but true.
        encoding = sys.getfilesystemencoding()
        # The string join is so that the shell is given the correct arguments:
        # this appears to be a subprocess bug.
        proc = self.ui.subprocess_Popen(' '.join(['gpgv', '--keyring',
            self.keyringpath.encode(encoding), sigpath.encode(encoding), '-']),
            stdin=subprocess.PIPE, stdout=subprocess.PIPE, stderr=subprocess.STDOUT,
            shell=True
            )
        out, err = proc.communicate(content)
        if proc.returncode != 0:
            self.ui.output_log(9, 'l_mirror.gpg', out)
            raise ValueError("GPG verification failed!")


class TestGPGVStrategy(object):
    """Test class for gpgv invocations.
    
    :ivar expected: The data that should be considered signed in a list, one
        item per call to verify.
    """
    
    def __init__(self, expected):
        self.expected = expected

    def verify(self, sigdir, signame, content):
        expected = self.expected.pop(0)
        if content != expected:
            raise ValueError("different data %r != %r" % (content, expected))



#
# LMirror is Copyright (C) 2010 Robert Collins <robertc@robertcollins.net>
# 
# LMirror is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
# 
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE.  See the GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <http://www.gnu.org/licenses/>.
# 
# In the LMirror source tree the file COPYING.txt contains the GNU General Public
# License version 3.
# 

"""Mirrorset, the core of l_mirror.

A MirrorSet is a node in a mirror network, and may be a root node if it has no
sources defined.

See the initialise function to create new MirrorSets, and MirrorSet for docs on
working with a MirrorSet.
"""

__all__ = ['initialise', 'MirrorSet']

import ConfigParser
from StringIO import StringIO
import time

from bzrlib import urlutils
from bzrlib.errors import NoSuchFile, NotLocalUrl

from l_mirror import gpg, journals


def initialise(base, name, content_root, ui):
    """Create a mirrorset at transport, called name, mirroring content_root.

    :param base: The directory under which to put the mirror set
        configuration and metadata. A bzrlib.transport.Transport.
    :param name: The name of the mirrorset.
    :param content_root: The root of the content to be mirrored. A
        bzrlib.transport.Transport.
    :param ui: A l_mirror.ui.AbstractUI.
    :return: A MirrorSet object.
    """
    setdir = base.clone('.lmirror/sets/%s' % name)
    ui.output_log(8, 'l_mirror.mirrorset', 'Creating mirrorset %s at %s'
        % (name, base.base))
    setdir.create_prefix()
    # Ideally we'd use O_EXCL|O_CREAT, but trying a manual open is close enough
    # to race-free, as if it does race, they'll both be creating the same
    # content.
    try:
        MirrorSet(base, name, ui)
    except ValueError:
        pass
    else:
        raise ValueError("Already a set %r configured at %s" % (name,
            base.base))
    setdir.put_bytes('format', '1\n')
    content_relative = urlutils.relative_url(base.base, content_root.base[:-1])
    if not content_relative:
        content_relative = '.'
    setdir.put_bytes('set.conf', '[set]\ncontent_root = %s\n' % content_relative)
    metadir = base.clone('.lmirror/metadata/%s' % name)
    metadir.create_prefix()
    metadir.put_bytes('format', '1\n')
    metadir.put_bytes('metadata.conf', '[metadata]\nbasis = 0\nlatest = 0\ntimestamp = 0\nupdating = True\n')
    journaldir = metadir.clone('journals')
    journaldir.create_prefix()
    journaldir.put_bytes('0', journals.Journal().as_bytes())
    return MirrorSet(base, name, ui)


class MirrorSet(object):
    """A mirrorable directory structure - a set of files to be mirrored.

    This is the primary model object in lmirror, using it one can manage mirror
    definitions, perform syncs etc.

    On disk a MirrorSet consists of configuration data in
    <basedir>/.lmirror/sets/<name>/:
    * format : Marker to allow compatibility.
    * set.conf: The configuration file. This shoud have one section:
      [set]
      content_root=<relpath to root>
    There is also state data in 
    <basedir>/.lmirror/metadata/<name>/:
    * format : Marker to allow compatibility.
    * metadata.conf: The current metadata. This has one section:
      [metadata]
      basis=<id of earliest full journal>
      latest=<id of latest journal>
      timestamp=<timestamp that last journal scan started at; controls what
        files are hashed and what just stated during a scan>
      updating=True|False # if True clients know that the mirror source may
        be out of sync with the metadata, and wait for that to get sorted.

    :ivar base: The base directory.
    :ivar name: The name of the mirror.
    :ivar excludes: () if not loaded from disk, or the exclude regexes to
        apply when scanning for changes.
    :ivar includes: () if not loaded from disk, or the include regexes to
        apply when scanning for changes.
    :ivar ui: The AbstractUI output is fed to.
    :ivar gpg_strategy: A bzrlib.gpg.GPGStrategy used for doing gpg signatures.
    :ivar gpgv_strategy: A l_mirror.gpg.GPGVStrategy for doing signature
        checking.
    """

    def __init__(self, base, name, ui):
        """Open an existing MirrorSet.

        :param base: A bzrlib.transport.Transport. The base directory where
            the mirror definition can be found.
        :param name: The name of the mirror set.
        :param ui: An l_mirror.ui.AbstractUI.
        """
        try:
            format = base.get_bytes('.lmirror/sets/%s/format' % name)
        except NoSuchFile:
            raise ValueError('No set %r - file not found' % name)
        if format != '1\n':
            raise ValueError('Unrecognised set format %r' % format)
        self.base = base
        self.name = name
        self.ui = ui
        self.excludes = ()
        self.includes = ()
        self.gpg_strategy = gpg.SimpleGPGStrategy(None)
        try:
            self.gpgv_strategy = gpg.GPGVStrategy(
                self._setdir().local_abspath('lmirror.gpg'), self.ui)
        except NotLocalUrl:
            # won't be able to do gpgv calls.
            self.gpgv_strategy = None

    def finish_change(self):
        """Scan the mirror set for changes and write a new journal entry.

        This will set updating=False and update the timestamp in the metadata.
        """
        metadata = self._get_metadata()
        if metadata.get('metadata', 'updating') != 'True':
            raise ValueError('No changeset open')
        self.ui.output_log(5, 'l_mirror.mirrorset',
            'finishing change of mirror set %s at %s' % 
            (self.name, self.base.base))
        last = float(metadata.get('metadata', 'timestamp'))
        now = time.time()
        basis = int(metadata.get('metadata', 'basis'))
        latest = int(metadata.get('metadata', 'latest'))
        current_state = self._combine_journals(basis, latest)
        updater = journals.DiskUpdater(current_state,
            self._content_root_dir(), self.name, last, self.ui,
            includes = self.get_includes(), excludes=self.get_excludes())
        journal = updater.finished()
        if journal.paths:
            next_id = latest + 1
            journal_bytes = journal.as_bytes()
            journal_dir = self._journaldir()
            journal_dir.put_bytes(str(next_id), journal_bytes)
            if self._is_signed():
                signature = self.gpg_strategy.sign(journal_bytes)
                journal_dir.put_bytes("%s.sig" % next_id, signature)
            metadata.set('metadata', 'latest', str(next_id))
            metadata.set('metadata', 'timestamp', str(now))
        else:
            self.ui.output_rest('No changes found in mirrorset.')
        metadata.set('metadata', 'updating', 'False')
        self._set_metadata(metadata)

    def get_excludes(self):
        if self.excludes == ():
            self._parse_content_conf()
        return self.excludes

    def get_includes(self):
        if self.includes == ():
            self._parse_content_conf()
        return self.includes

    def _is_signed(self):
        """Return true if the mirrorset is using gpg signatures."""
        return self._setdir().has('lmirror.gpg')

    def start_change(self):
        """Indicate to readers that changes are being made to the mirror.
        
        This simply sets updating=True in the metadata file.
        """
        metadata = self._get_metadata()
        if metadata.get('metadata', 'updating') != 'False':
            raise ValueError('Changeset already open')
        self.ui.output_log(5, 'l_mirror.mirrorset',
            'Marking mirror %s at %s as updating' %
            (self.name, self.base.base))
        metadata.set('metadata', 'updating', 'True')
        self._set_metadata(metadata)

    def cancel_change(self):
        """Cancel a scheduled change - simply unsets updating."""
        metadata = self._get_metadata()
        if metadata.get('metadata', 'updating') != 'True':
            raise ValueError('No changeset open')
        self.ui.output_log(5, 'l_mirror.mirrorset',
            'Marking mirror %s at %s as not updating' %
            (self.name, self.base.base))
        metadata.set('metadata', 'updating', 'False')
        self._set_metadata(metadata)

    def receive(self, another_mirrorset):
        """Perform a receive from another_mirrorset."""
        # XXX: check its a mirror of the same set. UUID or convergence?
        self.ui.output_log(5, 'l_mirror.mirrorset', 
            'Starting transmission from mirror %s at %s to %s at %s' %
            (another_mirrorset.name, another_mirrorset.base.base, self.name,
             self.base.base))
        metadata = self._get_metadata()
        source_meta = another_mirrorset._get_metadata()
        latest = int(metadata.get('metadata', 'latest'))
        source_latest = int(source_meta.get('metadata', 'latest'))
        signed = self._is_signed()       
        # XXX: BASIS: basis handling needed here (and thats when we
        # need non-overlapping syncing.
        if source_latest > latest:
            needed = range(latest + 1, source_latest + 1)
            new_journals = len(needed)
            combiner = journals.Combiner()
            source_journaldir = another_mirrorset._journaldir()
            journal_dir = self._journaldir()
            for journal_id in needed:
                journal_bytes = source_journaldir.get_bytes(str(journal_id))
                if signed:
                    # Copy the sig, check its valid in the current keyring.
                    sig_name = "%s.sig" % journal_id
                    sig_bytes = source_journaldir.get_bytes(sig_name)
                    journal_dir.put_bytes(sig_name, sig_bytes)
                    self.gpgv_strategy.verify(journal_dir, sig_name,
                        journal_bytes)
                journal_dir.put_bytes(str(journal_id), journal_bytes)
                journal = journals.parse(journal_bytes)
                combiner.add(journal)
            changed_paths = len(combiner.journal.paths)
            # If the keyring was changed and we were not signed before, copy
            # the keyring and check that all signed journals validate.
            keyringpath = '.lmirror/sets/%s/lmirror.gpg' % self.name
            if keyringpath in combiner.journal.paths:
                minijournal = journals.Journal()
                minijournal.paths[keyringpath] = combiner.journal.paths[
                    keyringpath]
                replayer = journals.TransportReplay(minijournal,
                    another_mirrorset.base, self.base, self.ui)
                replayer.replay()
                for journal_id in needed:
                    try:
                        sig_name = "%s.sig" % journal_id
                        sig_bytes = source_journaldir.get_bytes(sig_name)
                        journal_dir.put_bytes(sig_name, sig_bytes)
                    except NoSuchFile:
                        continue
                    self.gpgv_strategy.verify(journal_dir, sig_name,
                        journal_dir.get_bytes(str(journal_id)))
            replayer = journals.TransportReplay(combiner.journal,
                another_mirrorset.base, self.base, self.ui)
            replayer.replay()
            metadata.set('metadata', 'latest', str(source_latest))
            metadata.set('metadata', 'timestamp',
                source_meta.get('metadata', 'timestamp'))
            self._set_metadata(metadata)
        else:
            changed_paths = 0
            new_journals = 0
        self.ui.output_log(5, 'l_mirror.mirrorset', 
            'Received %d path changes in %d journals from mirror %s at '
            ' %s to %s at %s' %
            (changed_paths, new_journals, another_mirrorset.name,
             another_mirrorset.base.base, self.name, self.base.base))

    def _combine_journals(self, start, stop):
        """Combine a number of journals to get a tree model."""
        model = journals.Combiner()
        journaldir = self._journaldir()
        for journal_id in range(start, stop + 1):
            model.add(journals.parse(journaldir.get_bytes(str(journal_id))))
        return model.as_tree()

    def _setdir(self):
        return self.base.clone('.lmirror/sets/%s' % self.name)

    def _parse_content_conf(self):
        t = self._setdir()
        includes = []
        excludes = []
        try:
            file_bytes = t.get_bytes('content.conf')
            for line in file_bytes.split('\n'):
                if not line:
                    continue
                if line.startswith('include '):
                    includes.append(line[8:])
                elif line.startswith('exclude '):
                    excludes.append(line[8:])
        except NoSuchFile:
            pass
        self.includes = includes
        self.excludes = excludes

    def content_root_path(self):
        return self._get_settings().get('set', 'content_root')

    def _content_root_dir(self):
        return self.base.clone(self.content_root_path())

    def _metadata(self):
        """Get the transport for metadata."""
        return self.base.clone('.lmirror/metadata/%s' % self.name)

    def _journaldir(self):
        """Get the transport for journals."""
        return self._metadata().clone('journals')

    def _set_metadata(self, metadata):
        """Update metadata on disk."""
        output = StringIO()
        metadata.write(output)
        self._metadata().put_bytes('metadata.conf', output.getvalue())

    def _get_metadata(self):
        """Get a config parser with the metadata contents in it."""
        parser = OrderedConfigParser()
        parser.readfp(self._metadata().get('metadata.conf'))
        return parser

    def _get_settings(self):
        """Get a config parser with the set config contents in it."""
        parser = OrderedConfigParser()
        parser.readfp(self._setdir().get('set.conf'))
        return parser


class OrderedConfigParser(ConfigParser.ConfigParser):
    """Make write() not dictionary-order based."""

    def write(self, fp):
        """Write an .ini-format representation of the configuration state."""
        if self._defaults:
            fp.write("[%s]\n" % DEFAULTSECT)
            for (key, value) in sorted(self._defaults.items()):
                fp.write("%s = %s\n" % (key, str(value).replace('\n', '\n\t')))
            fp.write("\n")
        for section in self._sections:
            fp.write("[%s]\n" % section)
            for (key, value) in sorted(self._sections[section].items()):
                if key != "__name__":
                    fp.write("%s = %s\n" %
                             (key, str(value).replace('\n', '\n\t')))
            fp.write("\n")



#
# LMirror is Copyright (C) 2010 Robert Collins <robertc@robertcollins.net>
# 
# LMirror is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
# 
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE.  See the GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <http://www.gnu.org/licenses/>.
# 
# In the LMirror source tree the file COPYING.txt contains the GNU General Public
# License version 3.
# 

"""Monkeypatch helper function for tests.

This should be moved to testtools or something, its very generic.
"""

def monkeypatch(name, new_value):
    """Replace name with new_value.

    :return: A callable which will restore the original value.
    """
    location, attribute = name.rsplit('.', 1)
    # Import, swallowing all errors as any element of location may be
    # a class or some such thing.
    try:
        __import__(location, {}, {})
    except ImportError:
        pass
    components = location.split('.')
    current = __import__(components[0], {}, {})
    for component in components[1:]:
        current = getattr(current, component)
    old_value = getattr(current, attribute)
    setattr(current, attribute, new_value)
    def restore():
        setattr(current, attribute, old_value)
    return restore



#
# LMirror is Copyright (C) 2010 Robert Collins <robertc@robertcollins.net>
# 
# LMirror is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
# 
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE.  See the GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <http://www.gnu.org/licenses/>.
# 
# In the LMirror source tree the file COPYING.txt contains the GNU General Public
# License version 3.
# 

"""testtools.matchers.Matcher style matchers to help test l_mirror."""

from testtools.matchers import Matcher, Mismatch

__all__ = ['MatchesException']


class MatchesException(Matcher):
    """Match an exc_info tuple against an exception."""

    def __init__(self, exception):
        """Create a MatchesException that will match exc_info's for exception.
        
        :param exception: An exception to check against an exc_info tuple. The
            traceback object is not inspected, only the type and arguments of
            the exception.
        """
        Matcher.__init__(self)
        self.expected = exception

    def match(self, other):
        if type(other) != tuple:
            return _StringMismatch('%r is not an exc_info tuple' % other)
        if not issubclass(other[0], type(self.expected)):
            return _StringMismatch('%r is not a %r' % (
                other[0], type(self.expected)))
        if other[1].args != self.expected.args:
            return _StringMismatch('%r has different arguments to %r.' % (
                other[1], self.expected))

    def __str__(self):
        return "MatchesException(%r)" % self.expected


class _StringMismatch(Mismatch):
    """Convenience mismatch for simply-calculated string descriptions."""

    def __init__(self, description):
        self.description = description

    def describe(self):
        return self.description



#
# LMirror is Copyright (C) 2010 Robert Collins <robertc@robertcollins.net>
# 
# LMirror is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
# 
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE.  See the GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <http://www.gnu.org/licenses/>.
# 
# In the LMirror source tree the file COPYING.txt contains the GNU General Public
# License version 3.
# 

"""An Argument that looks up a command object."""

__all__ = ['CommandArgument']

from l_mirror.arguments import AbstractArgument
from l_mirror import commands


class CustomError(ValueError):

    def __str__(self):
        return self.args[0]


class CommandArgument(AbstractArgument):
    """An argument that looks up a command."""

    def _parse_one(self, arg):
        try:
            return commands._find_command(arg)
        except KeyError:
            raise CustomError("Could not find command '%s'." % arg)



#
# LMirror is Copyright (C) 2010 Robert Collins <robertc@robertcollins.net>
# 
# LMirror is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
# 
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE.  See the GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <http://www.gnu.org/licenses/>.
# 
# In the LMirror source tree the file COPYING.txt contains the GNU General Public
# License version 3.
# 

"""A test resource to provide isolation for the logging module (sigh, globals)."""

__all__ = ['LoggingResourceManager']

import logging

from testresources import TestResourceManager

from l_mirror.tests.monkeypatch import monkeypatch


class OldState:

    def __init__(self, restore_functions):
        self.restore_functions = restore_functions

    def tearDown(self):
        for fn in self.restore_functions:
            fn()


class LoggingResourceManager(TestResourceManager):
    """A resource for testing logging module using code.

    This resource resets the global logging state around a test.
    """

    def __getattribute__(self, attr):
        if attr == '_dirty':
            return True
        return object.__getattribute__(self, attr)

    def make(self, dep_resources):
        new_root = logging.RootLogger(logging.WARNING)
        new_manager = logging.Manager(new_root)
        return OldState([monkeypatch('logging.root', new_root),
            monkeypatch('logging.Logger.root', new_root),
            monkeypatch('logging.Logger.manager', new_manager)])

    def isDirty(self):
        return True

    def clean(self, resource):
        resource.tearDown()



#
# LMirror is Copyright (C) 2010 Robert Collins <robertc@robertcollins.net>
# 
# LMirror is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
# 
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE.  See the GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <http://www.gnu.org/licenses/>.
# 
# In the LMirror source tree the file COPYING.txt contains the GNU General Public
# License version 3.
# 

"""Get help on a command."""

from l_mirror.arguments import command
from l_mirror.commands import Command

class help(Command):
    """Get help on a command."""

    args = [command.CommandArgument('command_name', min=0)]

    def run(self):
        if not self.ui.arguments['command_name']:
            help = """lmirror -- large scale mirroring 
https://launchpad.net/l_mirror/

lmirror commands -- list commands
lmirror help [command] -- help system
"""
        else:
            cmd = self.ui.arguments['command_name'][0]
            help = cmd.__doc__
        self.ui.output_rest(help)
        return 0



#
# LMirror is Copyright (C) 2010 Robert Collins <robertc@robertcollins.net>
# 
# LMirror is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
# 
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE.  See the GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <http://www.gnu.org/licenses/>.
# 
# In the LMirror source tree the file COPYING.txt contains the GNU General Public
# License version 3.
# 

"""Tests for logging support code."""

from StringIO import StringIO
import logging
import os.path
import time

from l_mirror import logging_support
from l_mirror.tests import ResourcedTestCase
from l_mirror.tests.logging_resource import LoggingResourceManager
from l_mirror.tests.stubpackage import TempDirResource


class TestLoggingSetup(ResourcedTestCase):

    resources = [('logging', LoggingResourceManager())]

    def test_configure_logging_sets_converter(self):
        out = StringIO()
        c_log, f_log, formatter = logging_support.configure_logging(out)
        self.assertEqual(c_log, logging.root.handlers[0])
        self.assertEqual(f_log, logging.root.handlers[1])
        self.assertEqual(None, c_log.formatter)
        self.assertEqual(formatter, f_log.formatter)
        self.assertEqual(time.gmtime, formatter.converter)
        self.assertEqual("%Y-%m-%d %H:%M:%SZ", formatter.datefmt)
        self.assertEqual(logging.StreamHandler, c_log.__class__)
        self.assertEqual(out, c_log.stream)
        self.assertEqual(logging.FileHandler, f_log.__class__)
        self.assertEqual(os.path.expanduser("~/.cache/lmirror/log"), f_log.baseFilename)

    def test_can_supply_filename_None(self):
        out = StringIO()
        c_log, f_log, formatter = logging_support.configure_logging(out, None)
        self.assertEqual(None, f_log)



#
# LMirror is Copyright (C) 2010 Robert Collins <robertc@robertcollins.net>
# 
# LMirror is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
# 
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE.  See the GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <http://www.gnu.org/licenses/>.
# 
# In the LMirror source tree the file COPYING.txt contains the GNU General Public
# License version 3.
# 

"""Tests for the commands command."""

from l_mirror.commands import commands
from l_mirror.ui.model import UI
from l_mirror.tests import ResourcedTestCase


class TestCommandCommands(ResourcedTestCase):

    def get_test_ui_and_cmd(self):
        ui = UI()
        cmd = commands.commands(ui)
        ui.set_command(cmd)
        return ui, cmd

    def test_shows_a_table_of_commands(self):
        ui, cmd = self.get_test_ui_and_cmd()
        cmd.execute()
        self.assertEqual(1, len(ui.outputs))
        self.assertEqual('table', ui.outputs[0][0])
        self.assertEqual(('command', 'description'), ui.outputs[0][1][0])
        command_names = [row[0] for row in ui.outputs[0][1]]
        summaries = [row[1] for row in ui.outputs[0][1]]
        self.assertTrue('help' in command_names)
        self.assertTrue(
            'Get help on a command.' in summaries)



#!/usr/bin/env python
#
# LMirror is Copyright (C) 2010 Robert Collins <robertc@robertcollins.net>
# 
# LMirror is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
# 
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE.  See the GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <http://www.gnu.org/licenses/>.
# 
# In the LMirror source tree the file COPYING.txt contains the GNU General Public
# License version 3.
# 

from distutils.core import setup
import os

import l_mirror

version = '.'.join(str(component) for component in l_mirror.__version__[0:3])
phase = l_mirror.__version__[3]
if phase != 'final':
    import bzrlib.workingtree
    t = bzrlib.workingtree.WorkingTree.open_containing(__file__)[0]
    if phase == 'alpha':
        # No idea what the next version will be
        version = 'next-%s' % t.branch.revno()
    else:
        # Preserve the version number but give it a revno prefix
        version = version + '~%s' % t.branch.revno()

description = file(os.path.join(os.path.dirname(__file__), 'README.txt'), 'rb').read()

setup(name='lmirror',
      author='Robert Collins',
      author_email='robertc@robertcollins.net',
      url='https://launchpad.net/lmirror',
      description='Large scale mirror protocol.',
      long_description=description,
      scripts=['lmirror'],
      version=version,
      packages=[
        'l_mirror',
        'l_mirror.arguments',
        'l_mirror.commands',
        'l_mirror.ui',
        'l_mirror.tests',
        'l_mirror.tests.arguments',
        'l_mirror.tests.commands',
        'l_mirror.tests.ui',
        ])



#
# LMirror is Copyright (C) 2010 Robert Collins <robertc@robertcollins.net>
# 
# LMirror is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
# 
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE.  See the GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <http://www.gnu.org/licenses/>.
# 
# In the LMirror source tree the file COPYING.txt contains the GNU General Public
# License version 3.
# 

"""Tests for the help command."""

from l_mirror.commands import help, commands
from l_mirror.ui.model import UI
from l_mirror.tests import ResourcedTestCase


class TestCommand(ResourcedTestCase):

    def get_test_ui_and_cmd(self,args=()):
        ui = UI(args=args)
        cmd = help.help(ui)
        ui.set_command(cmd)
        return ui, cmd

    def test_shows_rest_of__doc__(self):
        ui, cmd = self.get_test_ui_and_cmd(args=['commands'])
        cmd.execute()
        self.assertEqual([('rest', commands.commands.__doc__)], ui.outputs)

    def test_shows_general_help_with_no_args(self):
        ui, cmd = self.get_test_ui_and_cmd()
        self.assertEqual(0, cmd.execute())
        self.assertEqual(1, len(ui.outputs))
        self.assertEqual('rest', ui.outputs[0][0])



#
# LMirror is Copyright (C) 2010 Robert Collins <robertc@robertcollins.net>
# 
# LMirror is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
# 
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE.  See the GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <http://www.gnu.org/licenses/>.
# 
# In the LMirror source tree the file COPYING.txt contains the GNU General Public
# License version 3.
# 

"""A TestResource that provides a temporary homedir package."""

__all__ = ['StubHomeResource']

import os

from testresources import TestResource

from l_mirror.tests.stubpackage import TempDirResource


class StubHome(object):
    """A temporary home dir.
    
    :ivar tempdir: The directory containing the package dir.
    """

    def __init__(self, tempdir):
        self.old_home = os.environ.get('HOME', None)
        os.environ['HOME'] = tempdir

    def tearDown(self):
        if self.old_home is None:
            del os.environ['HOME']
        else:
            os.environ['HOME'] = self.old_home


class StubHomeResource(TestResource):
    
    def __init__(self):
        super(StubHomeResource, self).__init__()
        self.resources = [('tempdir', TempDirResource())]

    def make(self, dependency_resources):
        tempdir = dependency_resources['tempdir']
        result = StubHome(tempdir)
        return result

    def cleanup(self, res):
        res.tearDown()



#
# LMirror is Copyright (C) 2010 Robert Collins <robertc@robertcollins.net>
# 
# LMirror is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
# 
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE.  See the GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <http://www.gnu.org/licenses/>.
# 
# In the LMirror source tree the file COPYING.txt contains the GNU General Public
# License version 3.
# 

"""Journals are the individual items that l_mirror uses to synchronise changes.

When a mirror set is updated a new journal is written (see DiskUpdater for
instance), and when a node is receiving updates it receives all the journals it
is missing, and then the updates determined by combining all the journals
together.

The parse() function can parse a journal bytes to make a new Journal object.

The Combiner object can combine multiple journals together, and then either
generate the model of a tree on disk, or a new journal with redundant changes
eliminated.

DiskUpdater is a class to compare a memory 'tree' and a bzr transport and 
output a journal to update the tree to match the transport.

Various Replay objects can replay a journal. TransportReplay replays a journal
by reading the file content from a transport object.
"""

__all__ = ['parse', 'Combiner', 'Journal', 'DiskUpdater', 'TransportReplay']

import errno
import os
from hashlib import sha1 as sha
import re

from bzrlib import errors, osutils


class Combiner(object):
    """Combine multiple journals.
    
    :ivar journal: The combined journal that is being created.
    """

    def __init__(self):
        """Create a Combiner object."""
        self.journal = Journal()

    def add(self, journal):
        """Add journal to the combined journal.

        :raises ValueError: If the journal cannot be safely combined.
        """
        # paths to delete after iteration.
        pending_del_paths = []
        merged_content = {}
        for path, new_content in journal.paths.iteritems():
            old_content = self.journal.paths.get(path, None)
            if old_content is None:
                continue
            # resolve here
            old_action = old_content[0]
            new_action = new_content[0]
            if (old_action, new_action) == ('new', 'new'):
                raise ValueError('Attempt to add %r twice.' % path)
            elif (old_action, new_action) == ('new', 'del'):
                if old_content[1] != new_content[1]:
                    raise ValueError('Attempt to delete wrong content %r, %r' %
                        (old_content, new_content))
                pending_del_paths.append(path)
            elif (old_action, new_action) == ('new', 'replace'):
                if old_content[1] != new_content[1][0]:
                    raise ValueError('Attempt to replace wrong content %r, %r' %
                        (old_content, new_content))
                merged_content[path] = ('new', new_content[1][1])
            elif (old_action, new_action) == ('del', 'new'):
                merged_content[path] = ('replace', (old_content[1], new_content[1]))
            elif (old_action, new_action) == ('del', 'del'):
                raise ValueError('Attempt to delete %r twice.' % path)
            elif (old_action, new_action) == ('del', 'replace'):
                raise ValueError('Attempt to replace deleted path %r.' % path)
            elif (old_action, new_action) == ('replace', 'new'):
                raise ValueError('Attempt to add %r twice.' % path)
            elif (old_action, new_action) == ('replace', 'del'):
                if old_content[1][1] != new_content[1]:
                    raise ValueError('Attempt to delete wrong content %r, %r' %
                        (old_content, new_content))
                merged_content[path] = ('del', old_content[1][0])
            elif (old_action, new_action) == ('replace', 'replace'):
                if old_content[1][1] != new_content[1][0]:
                    raise ValueError('Attempt to replace wrong content %r, %r' %
                        (old_content, new_content))
                merged_content[path] = ('replace', (old_content[1][0], new_content[1][1]))
            else:
                raise ValueError("Unknown action pair %r" % (
                    (old_action, new_action),))
        # backdoor for speed - XXX may be premature. 
        self.journal.paths.update(journal.paths)
        self.journal.paths.update(merged_content)
        for path in pending_del_paths:
            del self.journal.paths[path]

    def as_tree(self):
        """Convert a from-null combined journal into a tree.

        The tree is represented as a simple dict (path -> content) where content
        is either another dict, for directories, or the kind_data for the path.
        
        :return: A dict representing the tree that the journal would create if
            replayed.
        :raises ValueError: If the journal contains any delete, replace actions
            or items with missing parents.
        """
        result = {}
        for path, (action, kind_data) in sorted(self.journal.paths.iteritems()):
            if action in ('del', 'replace'):
                raise ValueError(
                    'cannot generate a tree representation for a partial '
                    ' journal, path %r is not new.' % path)
            segments = path.split('/')
            cwd = result
            for segment in segments[:-1]:
                try:
                    cwd = cwd[segment]
                except KeyError:
                    raise ValueError('Missing parent dir for path %r' % path)
            if kind_data[0] == 'dir':
                cwd[segments[-1]] = {}
            else:
                cwd[segments[-1]] = kind_data
        return result


class DiskUpdater(object):
    """Create a journal based on local disk and a tree representation.

    You can get a tree representation by using a Combiner to combine several
    journals (including a full snapshot, or starting from empty).

    :ivar tree: The tree to compare with.
    :ivar transport: The transport to read disk data from.
    :ivar last_timestamp: The timestamp of the most recent journal: all files
        modified more than 3 seconds before this timestamp are assumed to be
        unchanged.
    :ivar ui: A ui object to send output to.
    :ivar journal: The journal being built up.
    :ivar name: The name of the mirror set the journal will be updating. Used
        to include the mirror definition.
    :ivar include_re: The compiled re that, when it matches, indicates a path
        should be included.
    :ivar exclude_re: The compield re that, when it matches, indicates a path
        should only be included if it also matches the include_re. Note that
        only the exact path is considered: if you have an exclude rule '^foo'
        and an include rule '^foo/bar', once the directory 'foo' is excluded,
        'foo/bar' will not be examined and thus wont end up included. To deal
        with cases like that, either use a negative lookahead instead - exclude
        '^foo/(?!bar(?$|/)', or exclude paths starting with foo, and include
        both foo and bar explicitly: exclude '^foo(?:$|/)' and include '^foo$',
        '^foo/bar(?$|/)'.
    """

    def __init__(self, tree, transport, name, last_timestamp, ui,
        excludes=(), includes=()):
        """Create a DiskUpdater.

        :param tree: The tree to compare with.
        :param transport: The transport to read disk data from.
        :param name: The mirror set name, used to include its config in the
            mirror definition.
        :param last_timestamp: The timestamp of the most recent journal: all
            files modified more than 3 seconds before this timestamp are
            assumed to be unchanged. 3 seconds is chosen because it is larger
            than the 2 second fuzz needed to deal with FAT file systems.
        :param ui: A ui object to send output to.
        :param excludes: An optional list of uncompiled regexes to include in
            the exclude_re.
        :param includes: An optional list of uncompiled regexes to include in
            the include_re.
        """
        self.tree = tree
        self.transport = transport
        self.name = name
        self.last_timestamp = last_timestamp
        self.ui = ui
        self.journal = Journal()
        includes = [r'(?:^|/)\.lmirror/sets(?:$|/%s(?:$|/))' % name
            ] + list(includes)
        self.include_re = re.compile(self._make_re_str(includes))
        excludes = [r'(?:^|/)\.lmirror/'] + list(excludes)
        self.exclude_re = re.compile(self._make_re_str(excludes))

    def _make_re_str(self, re_strs):
        re_strs = ['(?:%s)' % re_str for re_str in re_strs]
        return '|'.join(re_strs)

    def finished(self):
        """Return the journal obtained by scanning the disk."""
        pending = ['']
        while pending:
            dirname = pending.pop(-1)
            names = self.transport.list_dir(dirname)
            # NB: quadratic in lookup here due to presence in inner loop:
            # consider tuning.
            segments = dirname.split('/')
            cwd = self.tree
            for segment in segments:
                if not segment:
                    continue
                try:
                    cwd = cwd[segment]
                except KeyError:
                    # totally new directory - added to journal by the directory
                    # above.
                    cwd = {}
            tree_names = set(cwd)
            names = set(names)
            for name in tree_names - names:
                # deletes
                path = dirname and ('%s/%s' % (dirname, name)) or name
                old_kind_details = cwd[name]
                if type(old_kind_details) is dict:
                    self._gather_deleted_dir(path, cwd[name])
                    old_kind_details = ('dir',)
                self.journal.add(path, 'del',
                    old_kind_details)
            new_names = names - tree_names
            for name in names:
                path = dirname and ('%s/%s' % (dirname, name)) or name
                if self._skip_path(path):
                    continue
                statinfo = self.transport.stat(path)
                # Is it old enough to not check
                mtime = getattr(statinfo, 'st_mtime', 0)
                if self.last_timestamp - mtime > 3 and name not in new_names:
                    continue
                kind = osutils.file_kind_from_stat_mode(statinfo.st_mode)
                if kind == 'file':
                    f = self.transport.get(path)
                    try:
                        disk_size, disk_sha1 = osutils.size_sha_file(f)
                    finally:
                        f.close()
                    new_kind_details = (kind, disk_sha1, disk_size)
                elif kind == 'symlink':
                    new_kind_details = (kind, os.readlink(self.transport.local_abspath(path)))
                elif kind == 'directory':
                    new_kind_details = ('dir',)
                    pending.append(path)
                else:
                    raise ValueError('unknown kind %r for %r' % (kind, path))
                if name in new_names:
                    self.journal.add(path, 'new', new_kind_details)
                else:
                    old_kind_details = cwd[name]
                    if type(old_kind_details) is dict:
                        old_kind_details = ('dir',)
                    if old_kind_details != new_kind_details:
                        self.journal.add(path, 'replace', (old_kind_details,
                            new_kind_details))
        for path, (action, details) in self.journal.paths.iteritems():
            self.ui.output_log(4, __name__, 'Journalling action %s for %r' % (
                action, path))
        return self.journal

    def _gather_deleted_dir(self, path, dirdict):
        # List what the tree thought it had as deletes.
        pending = [(path, dirdict)]
        while pending:
            dirname, cwd = pending.pop(-1)
            for name, old_kind_details in cwd:
                path = dirname and ('%s/%s' % (dirname, name)) or name
                if type(old_kind_details) is dict:
                    pending.append((path, old_kind_details))
                    old_kind_details = ('dir',)
                self.journal.add(path, 'del', old_kind_details)

    def _skip_path(self, path):
        """Should path be skipped?"""
        if (path.endswith('.lmirror/metadata') or
            path.endswith('.lmirrortemp')):
            # metadata is transmitted by the act of fetching the
            # journal.
            return True
        if self.exclude_re.search(path) and not self.include_re.search(path):
            return True
        return False


class Journal(object):
    """A journal of changes to a file system.
    
    :ivar paths: The paths that the journal alters. A dict from path to 
        action, kind_data.
    """

    def __init__(self):
        """Create a Journal."""
        self.paths = {}

    def add(self, relpath, action, kind_data):
        """Add a path to the journal.
        
        :param relpath: The path to journal.
        :param action: One of new, del, replace.
        :param kind_data: The data for the thing being added/deleted/replaced.
            In the special case of replacement this should be a two-tuple of
            the old data and the new data.
        """
        if relpath in self.paths:
            raise ValueError('path %r is already in use.' % relpath)
        if action == 'replace':
            if len(kind_data) != 2 or type(kind_data[0]) is not tuple:
                raise ValueError(
                    'looks like only one kind_data in replace action: %r' %
                    (kind_data,))
        self.paths[relpath] = (action, kind_data)

    def as_bytes(self):
        """Return a byte representation of this journal.

        The representation can be parsed by l_mirror.journals.parse. The
        structure is a header ('l-mimrror-journal-1\n') followed by '\\0'
        delimited fields. These follow the sequence PATH, ACTION, KIND_DATA and
        mirror the parameters to ``add``.

        :return: A bytesequence.
        """
        order = sorted(self.paths.items())
        output = []
        for path, (action, kind_data) in order:
            output.append(path)
            output.append(action)
            if action == 'replace':
                output.extend(kind_data[0])
                if type(output[-1]) is int:
                    output[-1] = str(output[-1])
                output.extend(kind_data[1])
            else:
                output.extend(kind_data)
            if type(output[-1]) is int:
                output[-1] = str(output[-1])
        return 'l-mirror-journal-1\n' + '\0'.join(output)


def parse(a_bytestring):
    """Parse a_bytestring into a journal.
    
    :return: A Journal.
    """
    header = 'l-mirror-journal-1\n'
    if not a_bytestring.startswith(header):
        raise ValueError('Not a journal: missing header %r in %r' % (
            header, a_bytestring))
    content = a_bytestring[len(header):]
    tokens = content.split('\x00')
    result = Journal()
    pos = 0
    if tokens[-1] == '':
        del tokens[-1]
    def parse_kind_data(pos):
        kind = tokens[pos]
        pos += 1
        if kind == 'file':
            return (kind, tokens[pos], int(tokens[pos+1])), pos + 2
        elif kind == 'dir':
            return (kind,), pos
        elif kind == 'symlink':
            return (kind, tokens[pos]), pos + 1
        else:
            raise ValueError('unknown kind %r at token %d.' % (kind, pos))
    while pos < len(tokens):
        path = tokens[pos]
        pos += 1
        action = tokens[pos]
        pos += 1
        if action in ('new', 'del'):
            kind_data, pos = parse_kind_data(pos)
        elif action == 'replace':
            kind_data1, pos = parse_kind_data(pos)
            kind_data2, pos = parse_kind_data(pos)
            kind_data = kind_data1, kind_data2
        result.add(path, action, kind_data)
    return result


class TransportReplay(object):
    """Replay a journal reading content from a transport.

    The replay() method is the main interface.
    
    :ivar sourcedir: The transport to read from.
    :ivar contentdir: The transport to apply changes to.
    :ivar journal: The journal to apply.
    :ivar ui: A UI for reporting with.
    """

    def __init__(self, journal, sourcedir, contentdir, ui):
        """Create a TransportReplay for journal from sourcedir to contentdir.

        :param journal: The journal to replay.
        :param sourcedir: The transport to read from.
        :param contentdir: The transport to apply changes to.
        :param ui: The ui to use for reporting.
        """
        self.journal = journal
        self.sourcedir = sourcedir
        self.contentdir = contentdir
        self.ui = ui

    def replay(self):
        """Replay the journal."""
        adds = []
        deletes = []
        replaces = []
        for path, (action, content) in self.journal.paths.iteritems():
            if action == 'new':
                adds.append((path, content))
            elif action == 'del':
                deletes.append((path, content))
            elif action == 'replace':
                replaces.append((path, content[0], content[1]))
            else:
                raise ValueError('unknown action %r for %r' % (action, path))
        # Ordering /can/ be more complex than just adds/replace/deletes:
        # might have an add blocked on a replace making a dir above it. Likewise
        # A delete might have to happen before a replace when a dir becomes a 
        # file. When we do smarter things, we'll have to just warn that we may
        # not honour general goals/policy if the user has given us such a
        # transform.
        # For now, simplest thing possible - want to get the concept performance
        # evaluated.
        adds.sort()
        for path, content in adds:
            self.put_with_check(path, content)()
        replaces.sort(reverse=True)
        to_rename = []
        try:
            for path, _, new_content in replaces:
                # TODO: (again, to do with replacing files with dirs:)
                #       do not delay creating dirs needed for files below them.
                to_rename.append(self.put_with_check(path, new_content))
            for path, old_content, new_content in replaces:
                # TODO: we may want to warn or perhaps have a strict mode here.
                # e.g. handle already deleted things. This should become clear
                # when recovery mode is done.
                self.contentdir.delete(path)
        finally:
            for doit in to_rename:
                doit()
        # Children go first :)
        deletes.sort(reverse=True)
        for path,content in deletes:
            self.ui.output_log(4, __name__, 'Deleting %s %r' % (content[0], path))
            try:
                if content[0] != 'dir':
                    self.contentdir.delete(path)
                else:
                    self.contentdir.rmdir(path)
            except errors.NoSuchFile:
                # Already gone, ignore it.
                pass

    def ensure_dir(self, path):
        """Ensure that path is a dir.

        If the path exists and is not a dir, an error is raised.
        """
        try:
            self.contentdir.mkdir(path)
        except errors.FileExists:
            st = self.contentdir.stat(path)
            if osutils.file_kind_from_stat_mode(st.st_mode) != 'directory':
                raise ValueError('unexpected non-directory at %r' % path)

    def check_file(self, path, content):
        """Check if there is a file at path with content.

        :raises: ValueError if there a non-file at path.
        :return: True if there is a file present with the right content.
        """
        try:
            st = self.contentdir.stat(path)
            if osutils.file_kind_from_stat_mode(st.st_mode) != 'file':
                raise ValueError('unexpected non-file at %r' % path)
            f = self.contentdir.get(path)
            try:
                size, sha1 = osutils.size_sha_file(f)
            finally:
                f.close()
            return sha1 == content[1] and size == content[2]
        except errors.NoSuchFile:
            return False

    def ensure_file(self, tempname, path, content):
        """Ensure that there is a file with content content at path.

        :param tempname: The name of a temporary file with the needed content.
        """
        if self.check_file(path, content):
            # Note that this implies we copied a file we didn't - during
            # a complete resync - sure, but still not optimal bw use.
            self.contentdir.delete(tempname)
        self.contentdir.rename(tempname, path)

    def ensure_link(self, realpath, target):
        """Ensure that realpath is a link to target.
        
        An error is raised if something is in the way.
        """
        try:
            os.symlink(content[1], realpath)
        except OSerror, e:
            if e.errno == errno.EEXIST:
                st = os.lstat(realpath)
                if osutils.file_kind_from_stat_mode(st.st_mode) != 'symlink':
                    raise ValueError('unexpected non-symlink at %r' % realpath)
                os.unlink(realpath)
                os.symlink(content[1], realpath)
            else:
                raise

    def put_with_check(self, path, content):
        """Put a_file at path checking that as received it matches content.

        :param path: A relpath.
        :param content: A content description of a file.
        :return: A callable that will execute the rename-into-place - all the
            IO has been done before returning.
        """
        tempname = '%s.lmirrortemp' % path
        self.ui.output_log(4, __name__, 'Writing %s %r' % (content[0], path))
        if content[0] == 'dir':
            return lambda: self.ensure_dir(path)
        elif content[0] == 'symlink':
            realpath = self.sourcedir.local_abspath(path)
            return lambda: self.ensure_link(realpath, content[1])
        elif content[0] != 'file':
            raise ValueError('unknown kind %r for %r' % (content[0], path))
        # don't download content we don't need
        try:
            if self.check_file(path, content):
                return lambda:None
        except ValueError:
            pass
        a_file = self.sourcedir.get(path)
        source = _ShaFile(a_file)
        try:
            # FIXME: mode should be supplied from above, or use 0600 and chmod
            # later.
            stream = self.contentdir.open_write_stream(tempname, 0644)
            try:
                size = osutils.pumpfile(source, stream)
            finally:
                stream.close()
            # TODO: here is where we should check for a mirror-is-updating
            # case.
            if size != content[2] or source.sha1.hexdigest() != content[1]:
                self.contentdir.delete(tempname)
                raise ValueError(
                    'read incorrect content for %r, got sha %r wanted %r' % (
                    path, source.sha1.hexdigest(), content[1]))
        finally:
            a_file.close()
        return lambda: self.ensure_file(tempname, path, content)


class _ShaFile(object):
    """Pretend to be a file, calculating the sha and size.
    
    After reading from this file, access the sha1 and size attributes to
    get the sha and size.

    XXX: I'm sure this is a dup with something bzrlib or somewhere else. Find
    and reuse.

    :ivar sha1: A sha1 object.
    """

    def __init__(self, a_file):
        self.a_file = a_file
        self.sha1 = sha()

    def read(self, amount=None):
        result = self.a_file.read(amount)
        self.sha1.update(result)
        return result



#
# LMirror is Copyright (C) 2010 Robert Collins <robertc@robertcollins.net>
# 
# LMirror is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
# 
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE.  See the GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <http://www.gnu.org/licenses/>.
# 
# In the LMirror source tree the file COPYING.txt contains the GNU General Public
# License version 3.
# 

"""Tests for the stubhome test helper."""

import os.path

from l_mirror.tests import ResourcedTestCase
from l_mirror.tests.stubhome import (
    StubHomeResource,
    )
from l_mirror.tests.stubpackage import (
    TempDirResource,
    )


class TestStubHomeResource(ResourcedTestCase):

    def test_has_tempdir(self):
        resource = StubHomeResource()
        self.assertEqual(1, len(resource.resources))
        self.assertIsInstance(resource.resources[0][1], TempDirResource)

    def test_sets_HOME(self):
        resource = StubHomeResource()
        home = resource.getResource()
        self.addCleanup(resource.finishedWith, home)
        self.assertEqual(home.tempdir, os.environ['HOME'])
